# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for auth in PyOkera
#
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=protected-access

import os
import tempfile

import unittest
import pytest

from okera.integration import domino
from okera.tests import pycerebro_test_common as common

class AuthTest(unittest.TestCase):
    def test_domino_basic(self):
        with pytest.raises(Exception) as excinfo:
            ctx = domino.context()
        assert 'Could not find Domino token environment variable' in str(excinfo.value)

        with pytest.raises(Exception) as excinfo:
            os.environ['DOMINO_TOKEN_FILE'] = '/does/not/exist'
            ctx = domino.context()
        assert 'Could not find Domino token file at' in str(excinfo.value)

        os.environ['DOMINO_TOKEN_FILE'] = os.path.join(
            os.environ.get('OKERA_HOME'),
            'integration', 'tokens', 'testuser_with_custom_claim.token'
        )
        ctx = domino.context()
        assert ctx._get_user() == 'testuser_preferred'

        with tempfile.NamedTemporaryFile() as fp:
            token_path = fp.name
            fp.write(b'customuser')
            fp.flush()

            os.environ['DOMINO_TOKEN_FILE'] = token_path
            ctx = domino.context()
            with common.get_planner(ctx) as conn:
                res = conn.scan_as_json('select * from okera_sample.whoami')
                assert len(res) == 1
                assert res[0]['user'] == 'customuser'

            fp.close()



# Copyright 2020 Okera Inc. All Rights Reserved.
#
# Some integration tests for auth in PyOkera
#
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=no-else-return
# pylint: disable=duplicate-code

import unittest

#from okera import context, _thrift_api
#from datetime import datetime
from okera.tests import pycerebro_test_common as common
from okera._thrift_api import (TAttribute, TAttributeMatchLevel, TGetDatasetsParams,
    TRecordServiceException)
from okera import odas

TEST_ROLE = "test_connection_filter_role"
TEST_USER = "test_connection_filter_user"
ROOT_USER = "root"
PSQL_CXN = 'test_connection_filter_psql_connection'
MYSQL_CXN = 'test_connection_filter_mysql_connection'
DB1 = 'test_connection_filter_db1'
DB2 = 'test_connection_filter_db2'
TBL1 = 'all_types'
TBL2 = 'scan_key_test'
TBL3 = 'all_numeric_types'
TBL4 = 'filter_pushdown_test'
NAMESPACE = 'test_namespace'
TAG1 = 'test_tag_1'
TAG2 = 'test_tag_2'
CREATE_PSQL_CXN = """
        CREATE DATACONNECTION %s CXNPROPERTIES
        ('connection_type'='JDBC',
        'jdbc_driver'='postgresql',
        'host'='jdbcpsqltest.cyn8yfvyuugz.us-west-2.rds.amazonaws.com',
        'port'='5432',
        'user_key'='awssm://postgres-jdbcpsqltest-username',
        'password_key'='awssm://postgres-jdbcpsqltest-password',
        'jdbc.db.name'='jdbc_test',
        'jdbc.schema.name'='public')
    """ % PSQL_CXN
CREATE_MYSQL_CXN = """
        CREATE DATACONNECTION %s CXNPROPERTIES
        ('connection_type'='JDBC',
        'jdbc_driver'='mysql',
        'host'='cerebro-db-test-long-running.cyn8yfvyuugz.us-west-2.rds.amazonaws.com',
        'port'='3306',
        'user_key'='awsps:///mysql/username',
        'password_key'='awsps:///mysql/password',
        'jdbc.db.name'='jdbc_test')
    """ % MYSQL_CXN

def get_table_names(result):
    return ["%s.%s" % (t.db[0], t.name) for t in result.datasets]

class ListDatasetsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """ Initializes one time state that is shared across test cases. This is used
            to speed up the tests. State that can be shared across (but still stable)
            should be here instead of __cleanup()."""
        super(ListDatasetsTest, cls).setUpClass()
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % DB1)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % DB2)
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS %s' % PSQL_CXN)
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS %s' % MYSQL_CXN)
            conn.execute_ddl(CREATE_PSQL_CXN)
            conn.execute_ddl(CREATE_MYSQL_CXN)

            # create DB1 using PSQL_CXN and load TBL1 in it using load_definitions
            conn.execute_ddl("""CREATE DATABASE %s DBPROPERTIES(
                'okera.connection.name' = '%s')""" % (DB1, PSQL_CXN))
            conn.execute_ddl("""ALTER DATABASE %s LOAD DEFINITIONS(%s, %s)""" \
                % (DB1, TBL1, TBL2))

            # create DB2 and manually create TBL2 with PSQL_CXN and TBL3 with MYSQL_CXN
            conn.execute_ddl('CREATE DATABASE %s' % DB2)
            conn.execute_ddl("""
                CREATE EXTERNAL TABLE %s.%s
                STORED as JDBC
                TBLPROPERTIES(
                'driver' = 'postgresql',
                'okera.connection.name' = '%s',
                'jdbc.db.name'='jdbc_test',
                'jdbc.schema.name'='public',
                'table' = '%s'
                )""" % (DB2, TBL3, PSQL_CXN, TBL3))
            conn.execute_ddl("""
                CREATE EXTERNAL TABLE %s.%s
                STORED as JDBC
                TBLPROPERTIES(
                'driver' = 'mysql',
                'okera.connection.name' = '%s',
                'jdbc.db.name'='jdbc_test',
                'jdbc.schema.name'='public',
                'table' = '%s'
                )""" % (DB2, TBL4, MYSQL_CXN, TBL4))

            # create test role and assign it to a group
            conn.execute_ddl('DROP ROLE IF EXISTS %s'% TEST_ROLE)
            conn.execute_ddl('CREATE ROLE %s'% TEST_ROLE)
            conn.execute_ddl('GRANT ROLE %s TO GROUP %s' % (TEST_ROLE, TEST_USER))

            # grant 'ALL' to the test role on DB2
            conn.execute_ddl('GRANT ALL ON DATABASE %s TO ROLE %s' % (DB2, TEST_ROLE))

            # create attributes
            conn.create_attribute(NAMESPACE, TAG1)
            conn.create_attribute(NAMESPACE, TAG2)

            # assign attributes at table level
            conn.assign_attribute(NAMESPACE, TAG1, DB1, TBL1)
            conn.assign_attribute(NAMESPACE, TAG2, DB1, TBL2)
            conn.assign_attribute(NAMESPACE, TAG1, DB2, TBL3)
            conn.assign_attribute(NAMESPACE, TAG2, DB2, TBL4)

    def _get_datasets(self, conn,
                      databases=None, dataset_names=None, filter=None,
                      tags=None, tag_match_level=None, connection_name=None,
                      offset=None, count=None):
        """ RPC to call GetDatasets()
        """
        request = TGetDatasetsParams()
        if conn.ctx._get_user():
            request.requesting_user = conn.ctx._get_user()
        request.databases = databases
        request.dataset_names = dataset_names
        request.filter = filter
        request.attributes = tags
        if tag_match_level:
            request.attr_match_level = TAttributeMatchLevel._NAMES_TO_VALUES[tag_match_level]
        request.connection_name = connection_name
        request.offset = offset
        request.count = count
        return conn._underlying_client().GetDatasets(request).datasets

    def _get_t_attribute_obj(self, namespace, key):
        attribute = TAttribute()
        attribute.attribute_namespace = namespace
        attribute.key = key
        return attribute

    def test_tag_filter(self):
        """ Test for multi-tag filter
        """
        db1 = 'attributes_test_db1'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # create attributes
            conn.create_attribute('test_namespace', 'test_key1')
            conn.create_attribute('test_namespace', 'test_key2')
            conn.create_attribute('test_namespace', 'test_key3')

            # create databases
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db1)
            conn.execute_ddl('CREATE DATABASE %s' % db1)

            # create tables
            conn.execute_ddl('CREATE TABLE %s.t1(c1 int, c11 int)' % db1)
            conn.execute_ddl('CREATE TABLE %s.t2(c2 int, c22 int)' % db1)
            conn.execute_ddl('CREATE TABLE %s.t3(c3 int, c33 int)' % db1)

            conn.assign_attribute('test_namespace', 'test_key1', db1)
            conn.assign_attribute('test_namespace', 'test_key2', db1, 't1')
            conn.assign_attribute('test_namespace', 'test_key3', db1, 't2', 'c2')

            # Create TAttribute objects required for testing
            attributes = []
            attr1_db_level = self._get_t_attribute_obj('test_namespace', 'test_key1')
            attr2_tab_level = self._get_t_attribute_obj('test_namespace', 'test_key2')
            attr3_col_level = self._get_t_attribute_obj('test_namespace', 'test_key3')

            # No Table should be returned.
            # Input tag is on DATABASE level and Match_Level = TABLE_ONLY
            attributes.append(attr1_db_level)
            datasets = conn.list_datasets(
                db1, tags=attributes, tag_match_level=TAttributeMatchLevel.TABLE_ONLY)
            self.assertTrue(len(datasets) == 0)

            # should return the table: t1
            # Input tag is on TABLE level and Match_Level = TABLE_ONLY
            attributes.clear()
            attributes.append(attr2_tab_level)
            datasets = conn.list_datasets(
                db1, tags=attributes, tag_match_level=TAttributeMatchLevel.TABLE_ONLY)
            self.assertTrue(len(datasets) == 1)
            for dataset in datasets:
                self.assertTrue(dataset.name == 't1')

            # No Table should be returned.
            # Input tag is on TABLE level and Match_Level = TABLE_ONLY
            attributes.clear()
            attributes.append(attr3_col_level)
            datasets = conn.list_datasets(
                db1, tags=attributes, tag_match_level=TAttributeMatchLevel.TABLE_ONLY)
            self.assertTrue(len(datasets) == 0)

            # should return the table: t1, t2
            # Input tags are on TABLE & COLUMN level and Match_Level = TABLE_PLUS
            attributes.clear()
            attributes.append(attr2_tab_level)
            attributes.append(attr3_col_level)
            datasets = conn.list_datasets(
                db1, tags=attributes, tag_match_level=TAttributeMatchLevel.TABLE_PLUS)
            self.assertTrue(len(datasets) == 2)
            returned_tbls = []
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue('t1' in returned_tbls)
            self.assertTrue('t2' in returned_tbls)

            # should return the table: t2
            # Input tags are on TABLE & COLUMN level and Match_Level = COLUMN_ONLY
            attributes.clear()
            attributes.append(attr2_tab_level)
            attributes.append(attr3_col_level)
            datasets = conn.list_datasets(
                db1, tags=attributes, tag_match_level=TAttributeMatchLevel.COLUMN_ONLY)
            self.assertTrue(len(datasets) == 1)
            for dataset in datasets:
                self.assertTrue(dataset.name == 't2')

            # No Table should be returned.
            # Input tags are on DB, TABLE & COLUMN level and Match_Level = DATABASE_ONLY
            attributes.clear()
            attributes.append(attr1_db_level)
            attributes.append(attr2_tab_level)
            attributes.append(attr3_col_level)
            datasets = conn.list_datasets(
                db1, tags=attributes, tag_match_level=TAttributeMatchLevel.DATABASE_ONLY)
            self.assertTrue(len(datasets) == 0)

            # should return the table: t1, t2, t3
            # Input tags are on DB, TABLE & COLUMN level and Match_Level = DATABASE_PLUS
            datasets = conn.list_datasets(
                db1, tags=attributes, tag_match_level=TAttributeMatchLevel.DATABASE_PLUS)
            self.assertTrue(len(datasets) == 3)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue('t1' in returned_tbls)
            self.assertTrue('t2' in returned_tbls)
            self.assertTrue('t3' in returned_tbls)

            # test to verify that if a tag matches at the DB level, all the tables
            # within that DB (after honoring other filters) should be returned
            # should return the table: t1, t2, t3
            # Input tags are only on DB level and Match_Level = DATABASE_PLUS
            attributes.clear()
            attributes.append(attr1_db_level)
            datasets = conn.list_datasets(
                db1, tags=attributes, tag_match_level=TAttributeMatchLevel.DATABASE_PLUS)
            self.assertTrue(len(datasets) == 3)
            returned_tbls = []
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue('t1' in returned_tbls)
            self.assertTrue('t2' in returned_tbls)
            self.assertTrue('t3' in returned_tbls)

            # drop databases
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db1)

    def test_connection_filter(self):
        """
            Test for connection_name filter only.
            No other filter would be input other than connection_name filter.
        """
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # DB1 has TBL1 which is associated with PSQL_CXN
            # DB1 has TBL2 which is associated with PSQL_CXN
            # DB2 has TBL3 which is associated with PSQL_CXN
            # DB2 has TBL4 which is associated with MYSQL_CXN
            # Test user has 'ALL' on DB2 but no access on DB1

            #  ********** Run tests as Root User ************
            ctx.enable_token_auth(token_str=ROOT_USER)
            # list datasets for connection_name = PSQL_CXN
            # TBL1, TBL2 and TBL3 should be returned as they are associated with PSQL_CXN
            datasets = self._get_datasets(conn, connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 3)
            returned_tbls = []
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)
            self.assertTrue(TBL2 in returned_tbls)
            self.assertTrue(TBL3 in returned_tbls)

            # list datasets for connection_name = MYSQL_CXN
            # only TLB4 should be returned as it is the only one associated with MYSQL_CXN
            datasets = self._get_datasets(conn, connection_name=MYSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL4 in returned_tbls)

            # list datasets for connection_name which does not exist
            # API should throw an AnalysisException
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                self._get_datasets(conn, connection_name='non_existing_connection')
            self.assertTrue('AnalysisException: Connection name does not exist.'
                in str(ex_ctx.exception))

            #  ********** Run tests as Test User ************
            ctx.enable_token_auth(token_str=TEST_USER)
            # list datasets for connection_name = PSQL_CXN (Test user)
            # Test user has full access on DB2 but none on DB1
            # only TBL3 should be returned as it is in DB2 and associated with PSQL_CXN
            # TBL1 & TBL2 should not be returned as they are in DB1 and user has no access on DB1
            datasets = self._get_datasets(conn, connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 not in returned_tbls)
            self.assertTrue(TBL2 not in returned_tbls)
            self.assertTrue(TBL3 in returned_tbls)

            # list datasets for connection_name = MYSQL_CXN (Test user)
            # Test user has full access on DB2 but none on DB1
            # only TBL4 should be returned as it is in DB2 and associated with MYSQL_CXN
            datasets = self._get_datasets(conn, connection_name=MYSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL4 in returned_tbls)

            # list datasets for connection_name which does not exist (Test user)
            # API should throw an AnalysisException
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                self._get_datasets(conn, connection_name='non_existing_connection')
            self.assertTrue('AnalysisException: Connection name does not exist.'
                in str(ex_ctx.exception))

    def test_connection_filter_should_not_return_internal_dbs(self):
        """
            Test for connection_name filter only.
            connection_filter should not return internal cralwer dbs.
        """
        INTERNAL_CRAWLER_DB = "_okera_crawler_connection_filter_test_db"
        TBL5 = "table_in_crawler_db"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # DB1 has TBL1 which is associated with PSQL_CXN
            # DB1 has TBL2 which is associated with PSQL_CXN
            # DB2 has TBL3 which is associated with PSQL_CXN
            # DB2 has TBL4 which is associated with MYSQL_CXN
            # Test user has 'ALL' on DB2 but no access on DB1

            #  ********** Run tests as Root User ************
            ctx.enable_token_auth(token_str=ROOT_USER)
            # create internal crawler db and manually create TBL5 with PSQL_CXN
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % INTERNAL_CRAWLER_DB)
            conn.execute_ddl('CREATE DATABASE %s' % INTERNAL_CRAWLER_DB)
            conn.execute_ddl("""
                CREATE EXTERNAL TABLE %s.%s
                STORED as JDBC
                TBLPROPERTIES(
                'driver' = 'postgresql',
                'okera.connection.name' = '%s',
                'jdbc.db.name'='jdbc_test',
                'jdbc.schema.name'='public',
                'table' = '%s'
                )""" % (INTERNAL_CRAWLER_DB, TBL5, PSQL_CXN, TBL1))

            # list datasets for connection_name = PSQL_CXN
            # TBL1, TBL2 and TBL3 should be returned as they are associated with PSQL_CXN
            # TBL5 should not be returned as it is in internal crawler DB.
            datasets = self._get_datasets(conn, connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 3)
            returned_tbls = []
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)
            self.assertTrue(TBL2 in returned_tbls)
            self.assertTrue(TBL3 in returned_tbls)
            self.assertTrue(TBL5 not in returned_tbls)

            # drop internal crawler db
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % INTERNAL_CRAWLER_DB)

    def test_conn_filter_with_databases_filter(self):
        """
            Test for connection_name filter in combination with database filter
        """
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # DB1 has TBL1 which is associated with PSQL_CXN
            # DB1 has TBL2 which is associated with PSQL_CXN
            # DB2 has TBL3 which is associated with PSQL_CXN
            # DB2 has TBL4 which is associated with MYSQL_CXN
            # Test user has 'ALL' on DB2 but no access on DB1

            #  ********** Run tests as Root User ************
            ctx.enable_token_auth(token_str=ROOT_USER)
            # list datasets for DB1 (without specifying connection_name filter)
            # TBL1 and TBL2 should be returned  (and not TBL3 and TBL4)
            datasets = self._get_datasets(conn, databases=[DB1])
            self.assertTrue(len(datasets) == 2)
            returned_tbls = []
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)
            self.assertTrue(TBL2 in returned_tbls)

            # list datasets for DB2 (without specifying connection_name filter)
            # TBL3 and TBL4 should be returned  (and not TBL1 and TBL2)
            datasets = self._get_datasets(conn, databases=[DB2])
            self.assertTrue(len(datasets) == 2)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL3 in returned_tbls)
            self.assertTrue(TBL4 in returned_tbls)

            # list datasets for database = DB1 and connection_name = PSQL_CXN
            # TBL1 - returned as it's in DB1 and associated with PSQL_CXN
            # TBL2 - returned as it's in DB1 and associated with PSQL_CXN
            # TBL3 - not returned even though it is with PSQL_CXN as it's not in DB1
            # TBL4 - not returned as it's neither in DB1 and nor associated with PSQL_CXN
            datasets = self._get_datasets(conn, databases=[DB1], connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 2)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)
            self.assertTrue(TBL2 in returned_tbls)

            # list datasets for database = DB1 and connection_name = MYSQL_CXN
            # TBL1 and TBL2 - not returned as they are not associated with MYSQL_CXN
            # TBL3 and TBL4 - not returned as they are not in DB1
            datasets = self._get_datasets(conn, databases=[DB1], connection_name=MYSQL_CXN)
            self.assertTrue(len(datasets) == 0)

            # list datasets for database = DB2 and connection_name = PSQL_CXN
            # TBL1 - not returned even though it is with PSQL_CXN as it's not in DB2
            # TBL2 - not returned even though it is with PSQL_CXN as it's not in DB2
            # TBL3 - returned as it's in DB2 and associated with PSQL_CXN
            # TBL4 - not returned as it's not associated with PSQL_CXN
            datasets = self._get_datasets(conn, databases=[DB2], connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL3 in returned_tbls)

            # list datasets for database = DB2 and connection_name = MYSQL_CXN
            # TBL1 - not returned as it's neither in DB2 and nor associated with MYSQL_CXN
            # TBL2 - not returned as it's neither in DB2 and nor associated with MYSQL_CXN
            # TBL3 - not returned as it's not associated with MYSQL_CXN
            # TBL4 - returned as it's in DB2 and associated with MYSQL_CXN
            datasets = self._get_datasets(conn, databases=[DB2], connection_name=MYSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL4 in returned_tbls)


            # list datasets for database = DB1,DB2 and connection_name = PSQL_CXN
            # TBL1 - returned as it's in DB1 and associated with PSQL_CXN
            # TBL2 - returned as it's in DB1 and associated with PSQL_CXN
            # TBL3 - returned as it's in DB2 and associated with PSQL_CXN
            # TBL4 - not returned as it's not associated with PSQL_CXN
            datasets = self._get_datasets(conn, databases=[DB1, DB2],
                                          connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 3)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)
            self.assertTrue(TBL2 in returned_tbls)
            self.assertTrue(TBL3 in returned_tbls)

            # list datasets for aatabase = DB1 and connection_name = 'non_existing_connection'
            # API should throw an AnalysisException
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                datasets = self._get_datasets(conn, databases=[DB1],
                                              connection_name='non_existing_connection')
            self.assertTrue('AnalysisException: Connection name does not exist.'
                in str(ex_ctx.exception))

            #  ********** Run tests as Test User ************
            ctx.enable_token_auth(token_str=TEST_USER)

            # list datasets for database = DB1 (test user)
            # Test user has no access on DB1
            # so No Table should be returned
            datasets = self._get_datasets(conn, databases=[DB1])
            self.assertTrue(len(datasets) == 0)

            # list datasets for database = DB2 (test user)
            # Test user has full access on DB2
            # TBL3 and TBL4 should be returned
            datasets = self._get_datasets(conn, databases=[DB2])
            self.assertTrue(len(datasets) == 2)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL3 in returned_tbls)
            self.assertTrue(TBL4 in returned_tbls)

            # list datasets for database = DB1 and connection_name = PSQL_CXN (test user)
            # Test user has no access on DB1
            # so No Table should be returned
            datasets = self._get_datasets(conn, databases=[DB1], connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 0)

            # list datasets for database = DB1 and connection_name = MYSQL_CXN (test user)
            # Test user has no access on DB1
            # so No Table should be returned
            datasets = self._get_datasets(conn, databases=[DB1], connection_name=MYSQL_CXN)
            self.assertTrue(len(datasets) == 0)

            # list datasets for database = DB2 and connection_name = PSQL_CXN (test user)
            # Test user has full access on DB2
            # only TBL3 should be returned as it is in DB2 and associated with PSQL_CXN
            datasets = self._get_datasets(conn, databases=[DB2], connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL3 in returned_tbls)

            # list datasets for database = DB2 and connection_name = MYSQL_CXN (test user)
            # Test user has full access on DB2
            # only TBL4 should be returned as it is in DB2 and associated with MYSQL_CXN
            datasets = self._get_datasets(conn, databases=[DB2], connection_name=MYSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL4 in returned_tbls)

            # list datasets for database = DB1,DB2 and connection_name = PSQL_CXN (test user)
            # Test user has no access on DB1
            # TBL1 - not returned as it's in DB1
            # TBL2 -  not returned as it's in DB1
            # TBL3 - returned as it's in DB2 and associated with PSQL_CXN
            # TBL4 - not returned as it's not associated with PSQL_CXN
            datasets = self._get_datasets(conn, databases=[DB1, DB2],
                                          connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL3 in returned_tbls)

            # list datasets for aatabase = DB1 and connection_name = 'non_existing_connection' (test user)
            # API should throw an AnalysisException
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                datasets = self._get_datasets(conn, databases=[DB1],
                                              connection_name='non_existing_connection')
            self.assertTrue('AnalysisException: Connection name does not exist.'
                in str(ex_ctx.exception))

            ctx.enable_token_auth(token_str=ROOT_USER)

    def test_conn_filter_with_pattern_filter(self):
        """
            Test for connection_name filter in combination with DBs and filter string
        """
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # DB1 has TBL1 which is associated with PSQL_CXN
            # DB1 has TBL2 which is associated with PSQL_CXN
            # DB2 has TBL3 which is associated with PSQL_CXN
            # DB2 has TBL4 which is associated with MYSQL_CXN
            # Test user has 'ALL' on DB2 but no access on DB1

            #  ********** Run tests as Root User ************
            ctx.enable_token_auth(token_str=ROOT_USER)
            # list datasets for DB1 and filter='types'
            # only TBL1 should be returned as it's in DB1 and its name matches the pattern
            datasets = self._get_datasets(conn, databases=[DB1], filter='types')
            self.assertTrue(len(datasets) == 1)
            returned_tbls = []
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)

            # list datasets for DB2 and filter='types'
            # only TBL3 should be returned as it's in DB2 and its name matches the pattern
            datasets = self._get_datasets(conn, databases=[DB2], filter='types')
            self.assertTrue(len(datasets) == 1)
            returned_tbls = []
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL3 in returned_tbls)

            # list datasets for databases=DB1, filter='types' and connection_name=PSQL_CXN
            # only TBL1 should be returned as it's in DB1 and its name matches the pattern
            # and it's also associated with PSQL_CXN
            datasets = self._get_datasets(conn, databases=[DB1], filter='types',
                                          connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls = []
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)

            # list datasets for databases=DB1, filter='types' and connection_name=MYSQL_CXN
            # No table should be returned none of them qualify for all three conditions
            datasets = self._get_datasets(conn, databases=[DB1], filter='types',
                                          connection_name=MYSQL_CXN)
            self.assertTrue(len(datasets) == 0)

            # list datasets for DB2 and filter='numeric' and connection_name=MYSQL_CXN
            # only TBL4 should be returned as it qualifies for all three conditions
            datasets = self._get_datasets(conn, databases=[DB2], filter='test',
                                          connection_name=MYSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls = []
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL4 in returned_tbls)

            # list datasets for DB2 and filter='test' and connection_name=PSQL_CXN
            # No table should be returned none of them qualify for all three conditions
            datasets = self._get_datasets(conn, databases=[DB2], filter='test',
                                          connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 0)

            # list datasets for DB1 & DB2 and filter='types'
            # TBL1 should be returned as it's in DB1 and its name matches the pattern
            # TBL3 should be returned as it's in DB2 and its name matches the pattern
            datasets = self._get_datasets(conn, databases=[DB1,DB2], filter='test')
            self.assertTrue(len(datasets) == 2)
            returned_tbls = []
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL2 in returned_tbls)
            self.assertTrue(TBL4 in returned_tbls)

            # list datasets for DB1 & DB2 and filter='non_existing_pattern'
            # No Table sould be returned due to failure of filter condition
            datasets = self._get_datasets(conn, databases=[DB1,DB2],
                                          filter='non_existing_pattern')
            self.assertTrue(len(datasets) == 0)

            # list datasets for DB1 & DB2 and filter='test' and connection_name=PSQL_CXN
            # only TBL2 should be returned as it qualifies for all three conditions
            datasets = self._get_datasets(conn, databases=[DB1,DB2], filter='test',
                                          connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls = []
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL2 in returned_tbls)

            # list datasets for DB1 & DB2 and filter='test' and connection_name=MYSQL_CXN
            # only TBL4 should be returned as it qualifies for all three conditions
            datasets = self._get_datasets(conn, databases=[DB1,DB2], filter='test',
                                          connection_name=MYSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls = []
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL4 in returned_tbls)

            # list datasets for DB1 & DB2 and filter='non_existing_pattern' and connection_name=MYSQL_CXN
            # No Table sould be returned due to failure of filter condition
            datasets = self._get_datasets(conn, databases=[DB1,DB2],
                                          filter='non_existing_pattern',
                                          connection_name=MYSQL_CXN)
            self.assertTrue(len(datasets) == 0)

            # list datasets for DB1 & DB2 and filter='test' and connection_name='non_existing_connection'
            # API should throw an AnalysisException
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                datasets = self._get_datasets(conn, databases=[DB1,DB2],
                                              filter='test',
                                              connection_name='non_existing_connection')
            self.assertTrue('AnalysisException: Connection name does not exist.'
                in str(ex_ctx.exception))

    def test_conn_filter_with_tag_filter(self):
        """
            Test for connection_name filter in combination with tag filter.
            Scenarios with other filter combinations would also be tested.
        """
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # DB1 has TBL1 which is associated with PSQL_CXN
            # DB1 has TBL2 which is associated with PSQL_CXN
            # DB2 has TBL3 which is associated with PSQL_CXN
            # DB2 has TBL4 which is associated with MYSQL_CXN
            # DB1.TBL1 has TAG1
            # DB1.TBL2 has TAG2
            # DB2.TBL3 has TAG1
            # DB2.TBL4 has TAG2
            # Test user has 'ALL' on DB2 but no access on DB1

            # Create TAttribute objects required for testing
            T_TAG1 = self._get_t_attribute_obj(NAMESPACE, TAG1)
            T_TAG2 = self._get_t_attribute_obj(NAMESPACE, TAG2)
            T_NON_EXISTING_TAG = self._get_t_attribute_obj(NAMESPACE, 'non_existing_tag')

            #  ********** Run tests as Root User ************
            ctx.enable_token_auth(token_str=ROOT_USER)
            returned_tbls = []

            # ****** Test cases for "Tag filters" only
            # list datasets for tags=TAG1 & TAG2 and tag_match_level=DATABASE_PLUS
            # TBL1, TB2, TBL3 and TBL4 all of them should be returned
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='DATABASE_PLUS')
            self.assertTrue(len(datasets) == 4)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)
            self.assertTrue(TBL2 in returned_tbls)
            self.assertTrue(TBL3 in returned_tbls)
            self.assertTrue(TBL4 in returned_tbls)

            # list datasets for tags=TAG1 & TAG2 and tag_match_level=DATABASE_ONLY
            # No tables should be returned as both the tags are on TABLE level
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='DATABASE_ONLY')
            self.assertTrue(len(datasets) == 0)

            # list datasets for tags=TAG1 & TAG2 and tag_match_level=TABLE_PLUS
            # TBL1, TB2, TBL3 and TBL4 all of them should be returned
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='TABLE_PLUS')
            self.assertTrue(len(datasets) == 4)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)
            self.assertTrue(TBL2 in returned_tbls)
            self.assertTrue(TBL3 in returned_tbls)
            self.assertTrue(TBL4 in returned_tbls)

            # list datasets for tags=TAG1 & TAG2 and tag_match_level=TABLE_ONLY
            # TBL1, TB2, TBL3 and TBL4 all of them should be returned
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='TABLE_ONLY')
            self.assertTrue(len(datasets) == 4)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)
            self.assertTrue(TBL2 in returned_tbls)
            self.assertTrue(TBL3 in returned_tbls)
            self.assertTrue(TBL4 in returned_tbls)

            # list datasets for tags=TAG1 & TAG2 and tag_match_level=COLUMN_ONLY
            # No tables should be returned as both the tags are on TABLE level
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='COLUMN_ONLY')
            self.assertTrue(len(datasets) == 0)

            # ********* Test cases for 'connection filter' with 'Tag filter' *******

            # list datasets for tags=TAG1 & TAG2 and connection_name=PSQL_CXN
            # TBL1, TB2 & TBL3 should be returned as they qualify for the conditions
            # TBL4 should not be returned as it's not associated with PSQL_CXN
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='TABLE_ONLY',
                                          connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 3)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)
            self.assertTrue(TBL2 in returned_tbls)
            self.assertTrue(TBL3 in returned_tbls)


            # list datasets for tags=TAG1 & TAG2 and connection_name=PSQL_CXN
            # No tables should be returned as both the tags are on TABLE level
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='DATABASE_ONLY',
                                          connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 0)

            # list datasets for tags=TAG2 and connection_name=MYSQL_CXN
            # only TBL4 should be returned as they qualify for the conditions
            # TBL1 & TBL3 should not be returned as they do not have TAG2
            # TBL2 should not be returned as it's not associated with MYSQL_CXN
            datasets = self._get_datasets(conn, tags=[T_TAG2],
                                          tag_match_level='TABLE_ONLY',
                                          connection_name=MYSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL4 in returned_tbls)

            # list datasets for tags=TAG1 and connection_name=PSQL_CXN
            # TBL1 & TBL3 should be returned as they qualify for the conditions
            # TBL2 should not be returned as it does not have TAG1
            # TBL4 should not be returned as it's not associated with PSQL_CXN
            datasets = self._get_datasets(conn, tags=[T_TAG1],
                                          tag_match_level='TABLE_ONLY',
                                          connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 2)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)
            self.assertTrue(TBL3 in returned_tbls)

            # list datasets for tags=NON_EXISTING_TAG and connection_name=PSQL_CXN
            # No Table should be returned
            datasets = self._get_datasets(conn, tags=[T_NON_EXISTING_TAG],
                                          tag_match_level='TABLE_ONLY',
                                          connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 0)

            # list datasets for tags=TAG1 & TAG2 and connection_name='non_existing_connection'
            # API should throw an AnalysisException
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                              tag_match_level='TABLE_ONLY',
                                              connection_name='non_existing_connection')
            self.assertTrue('AnalysisException: Connection name does not exist.'
                in str(ex_ctx.exception))

    def test_multiple_filter_combinations(self):
        """
            Test for multiple combinations of filters.
        """
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # DB1 has TBL1 which is associated with PSQL_CXN
            # DB1 has TBL2 which is associated with PSQL_CXN
            # DB2 has TBL3 which is associated with PSQL_CXN
            # DB2 has TBL4 which is associated with MYSQL_CXN
            # DB1.TBL1 has TAG1
            # DB1.TBL2 has TAG2
            # DB2.TBL3 has TAG1
            # DB2.TBL4 has TAG2
            # Test user has 'ALL' on DB2 but no access on DB1

            # Create TAttribute objects required for testing
            T_TAG1 = self._get_t_attribute_obj(NAMESPACE, TAG1)
            T_TAG2 = self._get_t_attribute_obj(NAMESPACE, TAG2)
            T_NON_EXISTING_TAG = self._get_t_attribute_obj(NAMESPACE, 'non_existing_tag')

            #  ********** Run tests as Root User ************
            ctx.enable_token_auth(token_str=ROOT_USER)
            returned_tbls = []

            ### Filter combination: tag, filter string, connection
            # list datasets for tags=TAG1 & TAG2, filter='test' & connection_name=PSQL_CXN
            # only TBL2 should be returned as it qualifies for all three conditions
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='TABLE_ONLY', filter='test',
                                          connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL2 in returned_tbls)

            ### Filter combination: tag, filter string, connection
            # list datasets for tags=TAG1 & TAG2, filter='types' & connection_name=PSQL_CXN
            # TBL1 & TBL3 should be returned as it qualifies for all three conditions
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='TABLE_ONLY', filter='types',
                                          connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 2)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)
            self.assertTrue(TBL3 in returned_tbls)

            ### Filter combination: database names, tag, filter string, connection
            # list datasets for databases =  DB1, tags = TAG1 & TAG2, filter='types'
            # and connection_name=PSQL_CXN
            # only TBL1 should be returned as it qualifies for all four conditions
            datasets = self._get_datasets(conn, databases=[DB1], tags=[T_TAG1,T_TAG2],
                                          tag_match_level='TABLE_ONLY', filter='types',
                                          connection_name=PSQL_CXN)
            self.assertTrue(len(datasets) == 1)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)

            ### Filter combination: dataset names and tags
            # list datasets for datasets =  TBL1, TBL2 & TBL3 and Tags = TAG1 & TAG2
            # TBL1 and TBL3 should be returned as they qualify for both conditions
            dataset_names = [(DB1 + '.' + TBL1), (DB1 + '.' + TBL2), (DB2 + '.' + TBL3)]
            datasets = self._get_datasets(conn, dataset_names=dataset_names,
                                          tags=[T_TAG1],
                                          tag_match_level='TABLE_ONLY')
            self.assertTrue(len(datasets) == 2)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)
            self.assertTrue(TBL3 in returned_tbls)

            ### Filter combination: dataset names and tags
            # list datasets for datasets = TBL1, TBL2 & TBL3 and Tags = T_NON_EXISTING_TAG
            # No Table qualifies for T_NON_EXISTING_TAG
            dataset_names = [(DB1 + '.' + TBL1), (DB1 + '.' + TBL2), (DB2 + '.' + TBL3)]
            datasets = self._get_datasets(conn, dataset_names=dataset_names,
                                          tags=[T_NON_EXISTING_TAG],
                                          tag_match_level='TABLE_ONLY')
            self.assertTrue(len(datasets) == 0)

    def test_invalid_filter_combinations(self):
        """ Test for verifying that certain filter combinations do not work. For example:
            - dataset_names and databases       can not be specified together
            - dataset_names and filter string   can not be specified together
            - dataset_names and connection_name can not be specified together
        """
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:

            dataset_names = [(DB1 + '.' + TBL1), (DB2 + '.' + TBL3)]
            databases = [DB1, DB2]
            filter = 'numeric'
            connection_name = PSQL_CXN

            # verifying dataset_names and databases filter combination
            # api should throw an AnalysisException
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                self._get_datasets(conn, dataset_names=dataset_names, databases=databases)
            self.assertTrue(
                'AnalysisException: Cannot specify both dataset_names and databases.'
                in str(ex_ctx.exception))

            # verifying dataset_names and filter string combination
            # api should throw an AnalysisException
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                self._get_datasets(conn, dataset_names=dataset_names, filter=filter)
            self.assertTrue(
                'AnalysisException: Cannot specify both dataset_names and filter.'
                in str(ex_ctx.exception))

            # verifying dataset_names and connection_name filter combination
            # api should throw an AnalysisException
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                self._get_datasets(conn, dataset_names=dataset_names, connection_name=connection_name)
            self.assertTrue(
                'AnalysisException: Cannot specify both dataset_names and connection_name.'
                in str(ex_ctx.exception))

    def test_pagination(self):
        """
            Test pagination with multiple filter combinations.
        """
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # DB1 has TBL1 which is associated with PSQL_CXN
            # DB1 has TBL2 which is associated with PSQL_CXN
            # DB2 has TBL3 which is associated with PSQL_CXN
            # DB2 has TBL4 which is associated with MYSQL_CXN
            # DB1.TBL1 has TAG1
            # DB1.TBL2 has TAG2
            # DB2.TBL3 has TAG1
            # DB2.TBL4 has TAG2
            # Test user has 'ALL' on DB2 but no access on DB1

            # Create TAttribute objects required for testing
            T_TAG1 = self._get_t_attribute_obj(NAMESPACE, TAG1)
            T_TAG2 = self._get_t_attribute_obj(NAMESPACE, TAG2)
            T_NON_EXISTING_TAG = self._get_t_attribute_obj(NAMESPACE, 'non_existing_tag')

            #  ********** Run tests as Root User ************
            ctx.enable_token_auth(token_str=ROOT_USER)
            returned_tbls = []

            # *********** Pagination with Tag filters *********
            # list datasets for tags=TAG1 & TAG2 and offset=0, count=2 (page 1)
            # TBL1, TB2 should be returned
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='TABLE_ONLY',
                                          offset=0, count=2)
            self.assertTrue(len(datasets) == 2)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)
            self.assertTrue(TBL2 in returned_tbls)

            # list datasets for tags=TAG1 & TAG2 and offset=2, count=2  (page 2)
            # TBL3, TB4 should be returned
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='TABLE_ONLY',
                                          offset=2, count=2)
            self.assertTrue(len(datasets) == 2)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL3 in returned_tbls)
            self.assertTrue(TBL4 in returned_tbls)

            # *********** Pagination with Tag and Connection filters *********
            # list datasets for tags=TAG1 & TAG2 , connection_name=PSQL_CXN
            # and offset=0, count=2  (page 1)
            # TBL1, TB2 should be returned
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='TABLE_ONLY',
                                          connection_name=PSQL_CXN,
                                          offset=0, count=2)
            self.assertTrue(len(datasets) == 2)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)
            self.assertTrue(TBL2 in returned_tbls)

            # list datasets for tags=TAG1 & TAG2 , connection_name=PSQL_CXN
            # and offset=2, count=2  (page 2)
            # only TBL3 should be returned
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='TABLE_ONLY',
                                          connection_name=PSQL_CXN,
                                          offset=2, count=2)
            self.assertTrue(len(datasets) == 1)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL3 in returned_tbls)

            # list datasets for tags=TAG1 & TAG2 , connection_name=PSQL_CXN
            # and offset=4, count=2  (page 3)
            # No Table should be returned
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='TABLE_ONLY',
                                          connection_name=PSQL_CXN,
                                          offset=4, count=2)
            self.assertTrue(len(datasets) == 0)

            # *********** Pagination with Tag and 'filter string' filters *********
            # list datasets for tags=TAG1 & TAG2 , filter='types'
            # and offset=0, count=2  (page 1)
            # only TBL1 should be returned
            # BUG: There is a bug here. Ideally, TBL1 and TBL3 should be returned as
            # both of them qualify for the conditions. However, we carry out the
            # pagination before applying the 'filter string', that's why result is incorrect.
            # TODO: Rethink and Rewrite the pagination logic
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='TABLE_ONLY',
                                          filter='types',
                                          offset=0, count=2)
            self.assertTrue(len(datasets) == 1)
            returned_tbls.clear()
            for dataset in datasets:
                returned_tbls.append(dataset.name)
            self.assertTrue(TBL1 in returned_tbls)

            # *********** Pagination with Tag and dataset_names filters *********
            # list datasets for tags=TAG1 & TAG2 , dataset_names=DB2.TBL3
            # and offset=0, count=2  (page 1)
            # No Table should be returned.
            # BUG: There is a bug here. Ideally, TBL3 should be returned as it qualifies
            # for the input conditions. However, we carry out the pagination twice
            # in case of dataset_names filter, that's why result is incorrect.
            # TODO: Rethink and Rewrite the pagination logic
            dataset_names = dataset_names = [DB2 + '.' + TBL3]
            datasets = self._get_datasets(conn, tags=[T_TAG1,T_TAG2],
                                          tag_match_level='TABLE_ONLY',
                                          dataset_names=dataset_names,
                                          offset=0, count=2)
            self.assertTrue(len(datasets) == 0)

    def test_list_datasets_is_public(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            role = "list_datasets_is_public_role"
            user = "list_datasets_is_public_role_user"
            db = "list_datasets_is_public_role_db"
            tbl1 = "%s.tbl1" % (db)
            tbl2 = "%s.tbl2" % (db)

            ctx.disable_auth()
            setup_ddls = [
                # Create a role and grant it to the user
                # This role will have an ABAC permission on a DB,
                # and the user should not see the tables in that DB as public
                "DROP ROLE IF EXISTS %s" % (role),
                "CREATE ROLE %s" % (role),
                "GRANT ROLE %s TO GROUP %s" % (role, user),

                # Create a DB and tables that we can grant on
                "DROP DATABASE IF EXISTS %s CASCADE" % (db),
                "CREATE DATABASE %s" % (db),
                "CREATE TABLE %s (col1 string)" % (tbl1),
                "CREATE TABLE %s (col1 string)" % (tbl2),

                # Grant an ABAC privilege on one of the tables
                "GRANT SELECT ON TABLE %s WHERE 1 = 1 TO ROLE %s" % (tbl1, role),

                # Grant the other table to the public role
                "GRANT SELECT ON TABLE %s TO ROLE %s" % (tbl2, "okera_public_role"),
            ]

            for ddl in setup_ddls:
                conn.execute_ddl(ddl)

            # Switch to the user
            ctx.enable_token_auth(token_str=user)

            # `tbl1` should not be public even though it has an ABAC
            # privilege on it.
            request = TGetDatasetsParams()
            request.dataset_names = [tbl1]
            request.requesting_user = user
            res = conn._underlying_client().GetDatasets(request)
            assert len(res.datasets) == 1
            assert len(res.is_public) == 1
            assert res.is_public[0] == False

            # `tbl2` should be public as it is granted to okera_public_role
            request = TGetDatasetsParams()
            request.dataset_names = [tbl2]
            request.requesting_user = user
            res = conn._underlying_client().GetDatasets(request)
            assert len(res.datasets) == 1
            assert len(res.is_public) == 1
            assert res.is_public[0] == True

    def test_list_datasets_with_db(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            role = "list_datasets_is_public_role"
            user = "list_datasets_is_public_role_user"
            db = "list_datasets_is_public_role_db1"
            tbl = "%s.tbl1" % (db)

            ctx.disable_auth()
            setup_ddls = [
                # Create a role and grant it to the user
                # This role will have an ABAC permission on a DB,
                # and the user should not see the tables in that DB as public
                "DROP ROLE IF EXISTS %s" % (role),
                "CREATE ROLE %s" % (role),
                "GRANT ROLE %s TO GROUP %s" % (role, user),

                # Create a DB and tables that we can grant on
                "DROP DATABASE IF EXISTS %s CASCADE" % (db),
                "CREATE DATABASE %s" % (db),
                "CREATE TABLE %s (col1 string)" % (tbl),

                "GRANT SELECT ON DATABASE %s TO ROLE %s" % (db, role),
            ]

            for ddl in setup_ddls:
                conn.execute_ddl(ddl)

            # Switch to the user
            ctx.enable_token_auth(token_str=user)

            tbl_name = tbl.split(".")[1]

            datasets = conn.list_datasets(db)
            assert len(datasets) == 1
            assert datasets[0].name == tbl_name and datasets[0].db[0] == db

            datasets = conn.list_datasets(db, filter=tbl_name)
            assert len(datasets) == 1
            assert datasets[0].name == tbl_name and datasets[0].db[0] == db

            datasets = conn.list_datasets(db, name=tbl_name)
            assert len(datasets) == 1
            assert datasets[0].name == tbl_name and datasets[0].db[0] == db

    def test_list_datasets_count_only(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            role = "list_datasets_count_only_role"
            user = "list_datasets_count_only_user"
            db1 = "list_datasets_count_only_db1"
            tbl11 = "%s.tbl1" % (db1)
            tbl12 = "%s.tbl2" % (db1)
            db2 = "list_datasets_count_only_db2"
            tbl21 = "%s.tbl1" % (db2)
            tbl22 = "%s.tbl2" % (db2)

            ctx.disable_auth()
            setup_ddls = [
                # Create a role and grant it to the user
                # This role will have an ABAC permission on a DB,
                # and the user should not see the tables in that DB as public
                "DROP ROLE IF EXISTS %s" % (role),
                "CREATE ROLE %s" % (role),
                "GRANT ROLE %s TO GROUP %s" % (role, user),

                # Create a DB and tables that we can grant on
                "DROP DATABASE IF EXISTS %s CASCADE" % (db1),
                "CREATE DATABASE %s" % (db1),
                "CREATE TABLE %s (col1 string)" % (tbl11),
                "CREATE TABLE %s (col1 string)" % (tbl12),
                "DROP DATABASE IF EXISTS %s CASCADE" % (db2),
                "CREATE DATABASE %s" % (db2),
                "CREATE TABLE %s (col1 string)" % (tbl21),
                "CREATE TABLE %s (col1 string)" % (tbl22),

                # Grant on the DB and tables
                "GRANT SELECT ON DATABASE %s TO ROLE %s" % (db1, role),
                "GRANT SELECT ON TABLE %s TO ROLE %s" % (tbl22, role),
            ]

            for ddl in setup_ddls:
                conn.execute_ddl(ddl)

            # Switch to the user
            ctx.enable_token_auth(token_str=user)

            # Verify the total count for our user is = 3 and
            # contains our tables
            request = TGetDatasetsParams()
            request.requesting_user = user
            request.databases = [db1, db2]
            request.count_only = True
            request.names_only = True
            res = conn._underlying_client().GetDatasets(request)
            tables = get_table_names(res)
            assert len(tables) == 3
            assert tbl11 in tables
            assert tbl12 in tables
            assert tbl21 not in tables
            assert tbl22 in tables

            # Verify it handles filters
            request = TGetDatasetsParams()
            request.requesting_user = user
            request.databases = [db1, db2]
            request.filter = 'tbl1'
            request.count_only = True
            request.names_only = True
            res = conn._underlying_client().GetDatasets(request)
            tables = get_table_names(res)
            assert len(tables) == 1
            assert tbl11 in tables
            assert tbl12 not in tables
            assert tbl21 not in tables
            assert tbl22 not in tables

            # Verify only name_only+count_only will work
            request = TGetDatasetsParams()
            request.requesting_user = user
            request.databases = [db1, db2]
            request.count_only = True
            request.names_only = False
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                res = conn._underlying_client().GetDatasets(request)
            self.assertTrue('Cannot specify count_only without names_only' in str(ex_ctx.exception))

            # Verify count_only + dataset names fails
            request = TGetDatasetsParams()
            request.requesting_user = user
            request.dataset_names = [tbl11]
            request.count_only = True
            request.names_only = True
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                res = conn._underlying_client().GetDatasets(request)
            self.assertTrue('Cannot specify both dataset_names and count_only' in str(ex_ctx.exception))

if __name__ == "__main__":
    unittest.main()



# Copyright 2020 Okera Inc. All Rights Reserved.
#
# Common methods for file format testing
#
# pylint: disable=line-too-long
# pylint: disable=too-many-locals
import json
import os
import subprocess
import shutil
import boto3
import requests

from okera.tests import pycerebro_test_common as common
from okera._thrift_api import TRecordServiceException

DEFAULT_PRESTO_PORT = int(os.environ['ODAS_TEST_PORT_PRESTO_COORD_HTTPS'])
DEFAULT_ENABLE_PRESTO = False
DEFAULT_ENABLE_SCAN_API = True
SCAN_API_RECORDS = 100

S3 = boto3.resource('s3')
TEST_BUCKET = 'cerebrodata-dev'

OKERA_HOME = os.environ["OKERA_HOME"]
OKERA_JAVA_VERSION = os.environ["OKERA_JAVA_VERSION"]
SCHEMA_TOOLS_JAR_NAME = OKERA_HOME + '/tools/schema-tools/target/schema-tools-' + \
    OKERA_JAVA_VERSION + '.jar'

class FileFormatTestBase(common.TestBase):

    @staticmethod
    def _get_test_iters(iters_default_value=None):
        # Run random generator tests
        if iters_default_value is not None:
            return iters_default_value
        if common.test_level_lt('checkin'):
            return 100
        if common.test_level_lt('ci'):
            return 1000
        return 5000

    @staticmethod
    def _get_test_unique_folder():
        # Use build_tag as unique identifier to avoid concurrent test run collisions.
        return os.getenv("BUILD_TAG", 'local')

    @staticmethod
    def _upload_s3(local_file_path, s3_file_path):
        s3_file_path = '%s/%s' % (FileFormatTestBase._get_test_unique_folder(),
                                  s3_file_path)
        S3.Bucket(TEST_BUCKET).upload_file(local_file_path, s3_file_path)
        return 's3://%s/%s' % (TEST_BUCKET, s3_file_path)

    @staticmethod
    def _write_s3(test_dir_prefix, path, data):
        test_dir_prefix = '%s/%s' % (FileFormatTestBase._get_test_unique_folder(),
                                     test_dir_prefix)
        filename = '%s/%s' % (test_dir_prefix, path)
        o = S3.Object(TEST_BUCKET, filename)
        o.put(Body=data)
        return 's3://%s/%s' % (TEST_BUCKET, filename)

    @staticmethod
    def _write_local_json_file(local_json_path, json_data):
        output = []
        for record in json_data:
            output.append(json.dumps(record))
        #print("output is: " + '\n'.join(output))
        with open(local_json_path, 'w') as outfile:
            outfile.write('\n'.join(output))

    @staticmethod
    def _read_local_json_file(local_json_path):
        expected_json = []
        with open(local_json_path, 'r') as data:
            for line in data:
                expected_json.append(json.loads(line))
        return expected_json

    @staticmethod
    def _generate_parquet(json_folder_path, parquet_folder_path):
        try:
            subprocess.check_output(['java', '-jar', SCHEMA_TOOLS_JAR_NAME, '--parquet',
                                     '--from-json', '--ignore-schema-errors',
                                     '--input-json-folder', json_folder_path,
                                     '--output-parquet-folder', parquet_folder_path],
                                    stderr=subprocess.STDOUT, timeout=300,
                                    universal_newlines=True)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output)

    @staticmethod
    def _generate_avro(avro_folder_path, test_schema_path, iters):
        try:
            subprocess.check_output(['java', '-jar', SCHEMA_TOOLS_JAR_NAME, '--avro',
                                     '--ignore-schema-errors',
                                     '--input-avro-schema', test_schema_path,
                                     '--output-avro-folder', avro_folder_path,
                                     '--sample-count', str(iters)],
                                    stderr=subprocess.STDOUT, timeout=300,
                                    universal_newlines=True)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output)

    @staticmethod
    def _recreate_dir(dir_name):
        try:
            shutil.rmtree(dir_name, ignore_errors=True)
            # put back empty folder, tests rely on the folder to exist.
            print('Creating directory: ' + dir_name)
            os.mkdir(dir_name)
        except OSError as e:
            print("Error: %s - %s." % (e.filename, e.strerror))

    @staticmethod
    def _cleanup(test_dir_prefix, test_name, local_dir='', test_dirs=None):
        bucket = S3.Bucket(TEST_BUCKET)
        test_dir_prefix = '%s/%s' % (FileFormatTestBase._get_test_unique_folder(),
                                     test_dir_prefix)
        prefix = '%s/%s' % (test_dir_prefix, test_name)
        bucket.objects.filter(Prefix=prefix).delete()
        if (local_dir != ''):
            test_dir = '%s/%s' % (local_dir, test_name)
            FileFormatTestBase._recreate_dir(test_dir)
            for test_dir in test_dirs:
                full_local_dir = '%s/%s/%s' % (local_dir, test_name, test_dir)
                FileFormatTestBase._recreate_dir(full_local_dir)

    def _test_path(self, path, expected_json, test_db, test_table,
                   allow_missing=False, batch_mode=False,
                   test_presto=DEFAULT_ENABLE_PRESTO, sql=None, required_type=None,
                   sort_by_record_idx=False, test_scan_api=DEFAULT_ENABLE_SCAN_API,
                   file_format='JSON'):
        """ Runs schema inference and verification for the json file at path. Returns
            (SQL, skipped) on failure/skip and None on success.
        """
        if not sql:
            sql = '''
      CREATE EXTERNAL TABLE %s LIKE %s '%s'
      STORED AS %s
      LOCATION '%s' ''' % (test_table, file_format, path, file_format, path)
        presto_port = None
        if test_presto:
            presto_port = DEFAULT_PRESTO_PORT
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=presto_port) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % test_db)
            conn.execute_ddl('CREATE DATABASE %s' % test_db)
            try:
                conn.execute_ddl(sql)
            except TRecordServiceException as ex:
                # Skip failures if the schema is longer than 4000
                if batch_mode and 'exceeds maximum type length' in ex.detail:
                    return sql, True
                raise
            if not batch_mode:
                print("SQL: %s" % sql)
            actual_schema = conn.plan(test_table).schema
            # Use max_task_count to make the results stable
            actual_json_pyokera = conn.scan_as_json(test_table, strings_as_utf8=True,
                                                    max_task_count=1,
                                                    dialect='okera')
            if presto_port:
                actual_json_presto = conn.scan_as_json(
                    'SELECT * FROM %s' % test_table, strings_as_utf8=True,
                    max_task_count=1, dialect='presto')

            if test_scan_api:
                url = 'http://%s:%s/scan/%s?records=%s' % \
                    (os.getenv('ODAS_TEST_HOST'), os.getenv('ODAS_TEST_PORT_PLANNER_UI'),\
                    test_table, SCAN_API_RECORDS)
                actual_scan_json = json.loads(requests.get(url).text)

            # Verify count with multiple tasks splits generated returns correct results.
            # This is done by passing the min_task_size=5K.
            # We had issues with boundary conditions around the json file splits.
            # This test will ensure the count is correct with and without multiple tasks.
            # FIXME: we have a bug with avro small file split size, we need to fix and
            # enable this for all datatypes.
            if file_format.lower() != 'avro':
                data = conn.scan_as_json(
                    'SELECT COUNT(*) AS count FROM %s ' % test_table,
                    strings_as_utf8=False, min_task_size=5 * 1000)
                sum = 0
                for cnt in data:
                    sum += cnt['count']
                # compare results against the presto data size.
                # We don't have to requery for count.
                self.assertEqual(sum, len(actual_json_pyokera))

        self.assertTrue(actual_schema is not None)
        # Verify scanning with pyokera
        res = FileFormatTestBase.compare_json(
            self, actual_json_pyokera, expected_json, allow_missing,
            empty_struct_equals_null=True, batch_mode=batch_mode,
            required_type=required_type, empty_array_equals_null=True)
        if not res:
            return sql, False

        # Verify presto if enabled
        if presto_port:
            if sort_by_record_idx:
                # Presto return records in a "random" order if the result set is
                # sufficiently large. Sort them by a key that exists to get comparable
                # results.
                actual_json_presto = sorted(actual_json_presto, key=lambda k: k['idx'])
                expected_json = sorted(expected_json, key=lambda k: k['idx'])
            res = FileFormatTestBase.compare_json(
                self, actual_json_presto, expected_json, allow_missing,
                empty_struct_equals_null=True, batch_mode=batch_mode,
                required_type=required_type,
                empty_array_equals_null=True)
            if not res:
                return sql, False

        if test_scan_api:
            res = FileFormatTestBase.compare_json(
                self, actual_scan_json, expected_json[0:SCAN_API_RECORDS],
                allow_missing, empty_struct_equals_null=True,
                batch_mode=batch_mode, required_type=required_type,
                empty_array_equals_null=True)
            if not res:
                return sql, False

        return None, False



# pylint: disable=bare-except
# pylint: disable=consider-using-enumerate
# pylint: disable=len-as-condition
# pylint: disable=protected-access
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-locals
# pylint: disable=unidiomatic-typecheck
# pylint: disable=bare-except

import datetime
import json
import math
import os
import random
import statistics
import string
import time
import unittest

import pytz

from okera import context, _thrift_api
from okera._thrift_api import TTypeId, TConfigType

DEFAULT_PRESTO_HOST = os.environ['ODAS_TEST_HOST']
DEFAULT_PRESTO_PORT = os.environ['ODAS_TEST_PORT_PRESTO_COORD_HTTPS']
DEFAULT_REST_SERVER_HOST = os.environ['ODAS_TEST_HOST']
DEFAULT_REST_SERVER_PORT = os.environ['ODAS_TEST_PORT_REST_SERVER']
DEFAULT_ACCESS_PROXY_HOST = os.environ['ODAS_TEST_HOST']
DEFAULT_ACCESS_PROXY_PORT = os.environ['ODAS_TEST_PORT_ACCESS_PROXY']

TEST_LEVELS = {
    'notests' : 0,
    'smoke' : 1,
    'dev' : 2,
    'checkin' : 3,
    'ci' : 4,
    'all' : 5,
}

TEST_LEVEL = 'smoke'
if 'TEST_LEVEL' in os.environ:
    TEST_LEVEL = os.environ['TEST_LEVEL']
elif 'CEREBRO_TEST_LEVEL' in os.environ:
    TEST_LEVEL = os.environ['CEREBRO_TEST_LEVEL']
ROOT_TOKEN = os.environ['OKERA_HOME'] + '/integration/tokens/cerebro.token'

identity = lambda x: x

def get_env_var(name, coercer, default):
    if name in os.environ:
        return coercer(os.environ[name])
    return default

def get_bool_env_var(name, default):
    return get_env_var(name, lambda x: str(x).lower() in ['true'], default)

def test_level_lt(min_level):
    if TEST_LEVEL not in TEST_LEVELS:
        return False
    return TEST_LEVELS[TEST_LEVEL] < TEST_LEVELS[min_level]

def get_test_context(auth_mech=None, dialect=None, namespace='okera', tz=pytz.utc):
    if auth_mech is None:
        auth_mech = get_env_var('PYCEREBRO_TEST_AUTH_MECH', identity, 'NOSASL')

    ctx = context(dialect=dialect, namespace=namespace, tz=tz)
    if auth_mech == 'NOSASL':
        ctx.disable_auth()
    elif auth_mech == 'TOKEN':
        ctx.enable_token_auth(token_file=ROOT_TOKEN)
    else:
        ctx.disable_auth()
    return ctx

def get_planner(ctx, host=None, port=None, dialect='okera', presto_port=None,
                namespace=None):
    if host is not None:
        host = host
    else:
        host = get_env_var('ODAS_TEST_HOST', identity, 'localhost')

    if port is not None:
        port = port
    else:
        port = get_env_var('ODAS_TEST_PORT_PLANNER_THRIFT', int, 12050)
    if 'presto' in dialect:
        ctx.enable_token_auth(token_str='root')
        return ctx.connect(host=host,
                           port=port,
                           presto_host=host,
                           presto_port=DEFAULT_PRESTO_PORT,
                           namespace=namespace)
    return ctx.connect(host=host, port=port, presto_port=presto_port, namespace=namespace)

def get_worker(ctx, host=None, port=None):
    if host is not None:
        host = host
    else:
        host = get_env_var('ODAS_TEST_HOST', identity, 'localhost')

    if port is not None:
        port = port
    else:
        port = get_env_var('ODAS_TEST_PORT_WORKER_THRIFT', int, 13050)

    return ctx.connect_worker(host=host, port=port)

def get_rest_server_url(scheme='http', host=DEFAULT_REST_SERVER_HOST, port=DEFAULT_REST_SERVER_PORT):
    return "%s://%s:%s" % (scheme, host, port)

def get_access_proxy_url(scheme='http', host=DEFAULT_ACCESS_PROXY_HOST, port=DEFAULT_ACCESS_PROXY_PORT):
    return "%s://%s:%s" % (scheme, host, port)

def measure_latency(iters, fn):
    durations = []
    for _ in range(0, iters):
        start = time.time()
        fn()
        durations.append((time.time() - start) * 1000)
    durations = sorted(durations)
    print("Iterations " + str(iters))
    print("Mean " + str(statistics.mean(durations)) + " ms")
    print("50%: " + str(durations[int(len(durations) * .5)]) + " ms")
    print("90%: " + str(durations[int(len(durations) * .90)]) + " ms")
    print("95%: " + str(durations[int(len(durations) * .95)]) + " ms")
    print("99%: " + str(durations[int(len(durations) * .99)]) + " ms")
    print("99.5%: " + str(durations[int(len(durations) * .995)]) + " ms")
    print("99.9%: " + str(durations[int(len(durations) * .999)]) + " ms")
    return durations

def configure_botocore_patch():
    os.environ['OKERA_PATCH_BOTO'] = 'True'
    os.environ['OKERA_PLANNER_HOST'] = \
        get_env_var('ODAS_TEST_HOST', identity, 'localhost')
    os.environ['OKERA_PLANNER_PORT'] = \
        get_env_var('ODAS_TEST_PORT_PLANNER_THRIFT', identity, 12050)
    from okera import initialize_default_context, check_and_patch_botocore
    initialize_default_context()
    check_and_patch_botocore()

def upsert_config(conn, config_type, key, value):
    """Upsert a configuration.

    config_type : TConfigType
        The type of configurations to return.
    key : str
        The key for this configuration. This must be unique with config_type.
    value : str
        The value for this config.

    Returns
    -------
    bool
        Returns true if the config was updated or false if it was inserted.
    list(str), optional
        List of warnings that were generated as part of this request.
    """
    request = _thrift_api.TConfigUpsertParams()
    request.config_type = config_type
    request.key = key
    if isinstance(value, str):
        request.params = {'value': value}
    else:
        request.params = value
    result = conn.service.client.UpsertConfig(request)
    # TODO: server needs to return if it was an upsert or not
    return True, result.warnings

def delete_config(conn, config_type, key):
    """Upsert a configuration.

    config_type : TConfigType
        The type of configurations to return.
    key : str
        The key for this configuration. This must be unique with config_type.

    Returns
    -------
    bool
        Returns true if the config was deleted.
    list(str), optional
        List of warnings that were generated as part of this request.
    """
    request = _thrift_api.TConfigDeleteParams()
    request.config_type = config_type
    request.key = key
    result = conn.service.client.DeleteConfig(request)
    # TODO: server needs to return if it was deleted or not
    return True, result.warnings

def list_configs(conn, config_type):
    """List configurations of the specified type.

    config_type : TConfigType
        The type of configurations to return.

    Returns
    -------
    map(str, str)
        List of configs as a map of key values.
    """
    table_name = None
    if config_type == TConfigType.AUTOTAGGER_REGEX:
        table_name = "okera_system.tagging_rules"
    elif config_type == TConfigType.SYSTEM_CONFIG:
        table_name = "okera_system.configs"
    else:
        raise ValueError("Invalid config type.")
    return conn.scan_as_json(table_name)

class SchemaNode():
    def __init__(self):
        self.col = None
        self.children = []

    @staticmethod
    def convert(cols, idx, schema):
        node = SchemaNode()
        node.col = cols[idx]
        schema.append(node)
        idx += 1
        if node.col.type.num_children:
            for _ in range(0, node.col.type.num_children):
                idx = SchemaNode.convert(cols, idx, node.children)
        return idx

class TmpView():
    def __init__(self, conn, sql):
        self.conn = conn
        # TODO: make this more unique so can be used concurrently
        self.db = "test_tmp_db"
        self.view = "tmp_view"
        self.conn.execute_ddl("CREATE DATABASE IF NOT EXISTS %s" % self.db)
        self.conn.execute_ddl("DROP VIEW IF EXISTS %s.%s" % (self.db, self.view))
        self.conn.execute_ddl("CREATE VIEW %s.%s AS %s" % (self.db, self.view, sql))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.conn.execute_ddl("DROP VIEW IF EXISTS %s.%s" % (self.db, self.view))

    def name(self):
        return '%s.%s' % (self.db, self.view)

class JsonGeneratorNode():
    def __init__(self):
        self.types = []
        self.children = {}

    def to_json(self):
        result = {}
        result['types'] = []
        for t in self.types:
            result['types'].append(TTypeId._VALUES_TO_NAMES[t])
        result['types'] = ', '.join(result['types'])
        if not self.children:
            return result
        for name, child in self.children.items():
            result[name] = child.to_json()
        return result

class JsonGenerator():
    def __init__(self,
                 types=None,
                 record_probabilities=None,
                 array_probabilties=None,
                 null_probability=.1,
                 empty_record_probability=.1,
                 min_fields=1,
                 max_fields=5,
                 max_array_len=3,
                 min_records=2,
                 max_records=10,
                 max_string_len=20,
                 seed=0,
                 max_recursion=5,
                 missing_fields_probability=0,
                 generate_variadic_schema=False,
                 generate_empty_record_all_types=False):
        random.seed(seed)
        if not types:
            types = [[TTypeId.BOOLEAN],
                     [TTypeId.BIGINT],
                     [TTypeId.DOUBLE],
                     [TTypeId.DATE],
                     [TTypeId.TIMESTAMP_NANOS],
                     [TTypeId.STRING]]
            if generate_variadic_schema:
                # This indicates that if this "type" is selected in the schema, the
                # data will be one of these types
                types.append([TTypeId.BIGINT, TTypeId.DOUBLE])
                types.append([TTypeId.BIGINT, TTypeId.DATE])
                types.append([TTypeId.DOUBLE, TTypeId.BOOLEAN])
                types.append([TTypeId.DATE, TTypeId.TIMESTAMP_NANOS])
                types.append([TTypeId.BOOLEAN, TTypeId.BIGINT, TTypeId.DATE,
                              TTypeId.TIMESTAMP_NANOS, TTypeId.DOUBLE, TTypeId.STRING])

        # Probabilities of generating record/array schemas by level. We want to
        # generate them with higher probabilities at the beginning to minimize
        # generating a lot of simple schemas.
        # This generate a CDF (i.e the remaining percentage is used to generate a
        # simple type).
        if not record_probabilities:
            record_probabilities = [.5, .4, .3, .25, .25, .2, .1]
        if not array_probabilties:
            array_probabilities = [.3, .3, .3, .25, .25, .2, .1]

        self.__min_fields = min_fields
        self.__max_fields = max_fields
        self.__min_records = min_records
        self.__max_records = max_records
        self.__max_array_len = max_array_len
        self.__null_probability = null_probability
        self.__max_string_len = max_string_len
        self.__empty_record_probability = empty_record_probability
        self.__max_recursion = max_recursion
        self.__types = types
        self.__array_probabilities = array_probabilities
        self.__record_probabilities = record_probabilities

        # Configs to control schema merge cases
        self.__generate_invalid_schema_merges = False
        self.__generate_empty_record_all_types = generate_empty_record_all_types
        self.__missing_fields_probability = missing_fields_probability

        self.__field_idx = 0
        self.__schema = None

        print("Generating with configuration")
        print("    Seed:  %s" % seed)
        print("    Types:  %s" % self.__types)
        print("    Record Probabilities:  %s" % self.__record_probabilities)
        print("    Array Probabilities:  %s" % self.__array_probabilities)
        print("    Max Array Len: %d" % (self.__max_array_len))
        print("    Max String Len: %d" % (self.__max_string_len))
        print("    Null Probability: %s" % (self.__null_probability))
        print("    Empty Record Probability: %s" % (self.__empty_record_probability))
        print("    Missing Field Probability: %s" % (self.__missing_fields_probability))
        print("    Max Depth: %d" % (self.__max_recursion))
        print("    # Fields: [%d, %d]" % (self.__min_fields, self.__max_fields))
        print("    # Records: [%d, %d]" % (self.__min_records, self.__max_records))
        print("    Generate variadic schemas: %s" % generate_variadic_schema)
        print("    Generate invalid merges: %s" %\
            self.__generate_invalid_schema_merges)
        print("    Generate empty records all types: %s" %\
            self.__generate_empty_record_all_types)

    def new_schema(self):
        self.__field_idx = 0
        self.__schema = self._generate_schema([TTypeId.RECORD], 0)

    def _random_type(self, level):
        prob_record = self.__record_probabilities[\
            min(level, len(self.__record_probabilities) - 1)]
        prob_array = self.__array_probabilities[\
            min(level, len(self.__array_probabilities) - 1)]
        r = random.random()
        if r < prob_record:
            return [TTypeId.RECORD]
        if r < prob_record + prob_array:
            return [TTypeId.ARRAY]
        return random.choice(self.__types)

    def _generate_schema(self, types, level):
        # Recursively generates a test schema
        node = JsonGeneratorNode()
        if level == self.__max_recursion:
            node.types = [TTypeId.STRING]
            return node

        node.types = types
        if types == [TTypeId.RECORD]:
            num_fields = random.randint(self.__min_fields, self.__max_fields - 1)
            for idx in range(0, num_fields):
                t = self._random_type(level)
                if level == 0:
                    name = 'c' + str(idx)
                else:
                    name = 'f' + str(self.__field_idx)
                    self.__field_idx += 1
                node.children[name] = self._generate_schema(t, level + 1)
        elif types == [TTypeId.ARRAY]:
            t = self._random_type(level)
            node.children['item'] = self._generate_schema(t, level + 1)
        return node

    def __generate_random_data(self, schema):
        if random.random() < self.__null_probability:
            return None
        t = random.choice(schema.types)
        if self.__generate_empty_record_all_types or t == TTypeId.RECORD:
            if random.random() < self.__empty_record_probability:
                return {}

        if t == TTypeId.BIGINT:
            digits = random.random() * 10
            v = int(pow(10, digits) * random.random())
            if random.random() < .1:
                return -v
            return v
        if t == TTypeId.DOUBLE:
            digits = random.random() * 10
            v = pow(10, digits) * random.random()
            if random.random() < .1:
                return -v
            return v
        if t == TTypeId.BOOLEAN:
            if random.random() > 0.5:
                return True
            return False
        if t == TTypeId.DATE:
            return '2020-01-01'
        if t == TTypeId.TIMESTAMP_NANOS:
            return '2020-01-01 01:02:03.123'
        if t == TTypeId.STRING:
            n = random.randint(0, self.__max_string_len - 1)
            return ''.join(random.choice(string.ascii_lowercase) for i in range(n))
        if t == TTypeId.RECORD:
            return self.__generate_record(schema)
        if t == TTypeId.ARRAY:
            return self.__generate_array(schema.children['item'])
        return 'Unsupported Type %s' % t

    def __generate_array(self, schema):
        result = []
        num_children = random.randint(0, self.__max_array_len - 1)
        for _ in range(0, num_children):
            result.append(self.__generate_random_data(schema))
        return result

    def __generate_record(self, schema):
        record = {}
        for name, child in schema.children.items():
            if random.random() < self.__missing_fields_probability:
                continue
            record[name] = self.__generate_random_data(child)
        return record

    def generate(self, generate_record_idx=False):
        n = self.__min_records
        if self.__min_records != self.__max_records:
            n = random.randint(self.__min_records, self.__max_records - 1)
        data = []
        for idx in range(0, n):
            r = self.__generate_record(self.__schema)
            if generate_record_idx:
                r['idx'] = idx
            data.append(r)
        return data

class TestBase(unittest.TestCase):
    """ Base class with some common test utilities. """
    @staticmethod
    def _ddl_count(conn, sql):
        return len(conn.execute_ddl(sql))

    @staticmethod
    def _recreate_test_role(conn, role, users):
        conn.execute_ddl("DROP ROLE IF EXISTS %s" % role)
        conn.execute_ddl("CREATE ROLE %s" % role)
        for user in users:
            conn.execute_ddl("GRANT ROLE %s TO GROUP %s" % (role, user))

    @staticmethod
    def _recreate_test_db(conn, db):
        conn.execute_ddl("DROP DATABASE IF EXISTS %s CASCADE" % db)
        conn.execute_ddl("CREATE DATABASE %s" % db)

    @staticmethod
    def get_random_leaf_column(conn, db, tbl):
        """ Returns a random path to a leaf col in this table."""
        tbl = conn.list_datasets(db, name=tbl)[0]
        schema = TestBase.convert_schema(tbl.schema.cols)
        result = ''
        flattened = ''
        while schema:
            col = random.choice(schema)
            if col.col.name:
                if result:
                    result += '.'
                if flattened:
                    flattened += '__'
                result += col.col.name.lower()
                flattened += col.col.name.lower()
            # else:
            #     if flattened:
            #         flattened += '__'
            #     flattened += 'item'
            schema = col.children
        return result, flattened

    @staticmethod
    def convert_schema(cols):
        """ Converts a flattened inorder schema list to a tree """
        schema = []
        idx = 0
        while idx < len(cols):
            idx = SchemaNode.convert(cols, idx, schema)
        return schema

    @staticmethod
    def _visible_cols(cols):
        result = []
        for c in cols:
            if c.hidden:
                continue
            result.append(c)
        return result

    @staticmethod
    def _top_level_columns(cols):
        total_children = 0
        for c in cols:
            if c.type.num_children:
                total_children += c.type.num_children
        return len(cols) - total_children

    @staticmethod
    def collect_column_attributes(ds):
        result = {}
        for col in ds.schema.cols:
            if not col.attribute_values:
                continue
            for v in col.attribute_values:
                key = v.database
                if v.table:
                    key += '.' + v.table
                    if v.column:
                        key += '.' + v.column
                if key not in result:
                    result[key] = []
                result[key].append(
                    v.attribute.attribute_namespace + '.' + v.attribute.key)
        for _, v in result.items():
            v.sort()
        return result


    @staticmethod
    def _try_parse_datetime(v):
        if not isinstance(v, str):
            return None

        FORMATS = [
            '%Y-%m-%d %H:%M:%S',
            '%Y-%m-%d %H:%M:%S.%f',
            '%Y-%m-%dT%H:%M:%S.%fZ',
            '%Y-%m-%dT%H:%M:%S.%f',
            '%Y-%m-%dT%H:%M:%SZ',
            '%Y-%m-%dT%H:%M:%S',
            '%Y-%m-%d',
        ]

        for fmt in FORMATS:
            try:
                return datetime.datetime.strptime(v, fmt)
            except ValueError:
                pass
        return None

    @staticmethod
    def _is_float(v):
        if v is None:
            return False
        try:
            float(v)
            return True
        except:
            return False

    @staticmethod
    def _equals(v1, v2):
        if TestBase._is_float(v1) and TestBase._is_float(v2):
            v1 = float(v1)
            v2 = float(v2)

        if isinstance(v1, float) and isinstance(v2, float):
            return v1 == v2 or (math.isnan(v1) and math.isnan(v2)) \
                or abs(v1 - v2) < .0001
        if v1 != v2:
            # Try as datetime with different formats
            d1 = TestBase._try_parse_datetime(v1)
            d2 = TestBase._try_parse_datetime(v2)
            if d1 is not None and d1 == d2:
                return True
            print("Values do not match. %s != %s" % (v1, v2))
            print("Types: %s != %s" % (type(v1), type(v2)))
        return v1 == v2

    @staticmethod
    def _is_empty_dictionary(v):
        """ Checks if v is None or a dictionary (recursively) of None values """
        if v is None:
            return True
        if type(v) != dict:
            return False
        for _, val in v.items():
            if not TestBase._is_empty_dictionary(val):
                return False
        return True

    @staticmethod
    def __deep_compare(actual, expected, allow_missing, empty_struct_equals_null,
                       required_type, empty_array_equals_null, zero_equals_none):
        if empty_struct_equals_null:
            if type(actual) == dict and (expected is None or len(expected) == 0):
                # We don't support nullable structs, so instead every field in the
                # struct is None
                for k, v in actual.items():
                    if not TestBase._is_empty_dictionary(v):
                        print("actual has a non-null struct. Expecting null.")
                        return False
                return True
            if type(expected) == dict and len(expected) == 0 and actual is None:
                # Allow empty struct to be considered equal to null. A struct with no
                # fields is otherwise not possible.
                return True

        # For parquet file format, the actual is empty for an empty list.
        if empty_array_equals_null:
            if type(expected) == list and (actual is None):
                return True

        # As we are strongly typed, we will convert the types to the "bigger" type for
        # schema merge cases.
        if type(actual) == str and type(expected) in [int, float]:
            expected = str(expected)
        if type(actual) == str and type(expected) == bool:
            if expected:
                expected = "true"
            else:
                expected = "false"
        if type(actual) == float and type(expected) == int:
            expected = float(expected)

        if type(actual) != type(expected):
            if type(actual) == list and type(expected) == dict:
                # Can't tell difference between empty list and empty dict
                if not actual and not expected:
                    return True
            # Handle some schema merge cases that are ambiguous with empty records
            if type(actual) == dict and not actual and \
                    type(expected) in [bool, float, int]:
                return True
            if type(actual) in [bool, float, int] and type(expected) == dict \
                    and not expected:
                return True
            if type(actual) == float and math.isnan(actual) and expected is None:
                return True
            if actual == "" and expected is None and zero_equals_none:
                return True
            if type(actual) == type(None) and type(expected) in [float, int]:
                if zero_equals_none and expected in [0, 0.0]:
                    return True
            if required_type and type(expected) in [bool, float, int, str]:
                if type(expected) != required_type:
                    # The required type doesn't match so the expected should be None
                    if actual is not None and \
                            not TestBase._equals(actual, required_type(expected)):
                        print("Expecting actual to be None. %s, %s" % (actual, expected))
                        return False
                    return True
            print("Types don't match %s != %s" % (type(actual), type(expected)))
            print("%s != %s" % (actual, expected))
            return False

        if type(actual) == dict:
            for k, v in actual.items():
                # Handle the case where some field that exists in the Okera version
                # but is null might not exist in the raw file version, but only if
                # allow_missing is set
                if k in expected:
                    if not TestBase.__deep_compare(
                            v, expected[k], allow_missing,
                            empty_struct_equals_null, required_type,
                            empty_array_equals_null, zero_equals_none):
                        print("Key %s from expected is not in actual." % k)
                        print("%s != %s" % (v, expected[k]))
                        return False
                elif k not in expected and not \
                        (TestBase._is_empty_dictionary(v) and allow_missing):
                    print("Key %s from actual is not in expected." % k)
                    return False
            for k, v in expected.items():
                if k not in actual and not allow_missing:
                    print("Key %s from expected is not in actual." % k)
                    return False
        elif type(actual) == list:
            if len(actual) != len(expected):
                print("Length of arrays are not equal. %s != %s" % \
                      (len(actual), len(expected)))
                return False
            for idx in range(len(actual)):
                if not TestBase.__deep_compare(
                        actual[idx], expected[idx], allow_missing,
                        empty_struct_equals_null, required_type,
                        empty_array_equals_null, zero_equals_none):
                    print("List elements don't match at idx %s. %s != %s" % \
                          (idx, actual[idx], expected[idx]))
                    return False
        else:
            return TestBase._equals(actual, expected)

        return True

    @staticmethod
    def _lower_keys(x):
        if isinstance(x, list):
            return [TestBase._lower_keys(v) for v in x]
        if isinstance(x, dict):
            return dict((k.lower(), TestBase._lower_keys(v)) for k, v in x.items())
        return x

    @staticmethod
    def compare_json(asserter, actual, expected, allow_missing, empty_struct_equals_null,
                     batch_mode, required_type, empty_array_equals_null,
                     zero_equals_none=False):
        actual = TestBase._lower_keys(actual)
        expected = TestBase._lower_keys(expected)

        if batch_mode and len(actual) != len(expected):
            print("Failure: Lengths did not match %s != %s" % \
                (len(actual), len(expected)))
            return False
        asserter.assertEqual(len(actual), len(expected))

        for idx in range(len(actual)):
            obj1 = actual[idx]
            obj2 = expected[idx]

            if TestBase.__deep_compare(
                    obj1, obj2, allow_missing, empty_struct_equals_null,
                    required_type, empty_array_equals_null, zero_equals_none):
                continue
            if batch_mode:
                return False
            print("EXPECTED:\n%s" % json.dumps(expected, indent=2, sort_keys=True))
            print("\nACTUAL:\n%s" % json.dumps(actual, indent=2, sort_keys=True))
            asserter.assertEqual(json.dumps(actual, indent=2, sort_keys=True),
                                 json.dumps(expected, indent=2, sort_keys=True))
        return True

def random_string(length):
    return ''.join(random.choice(string.ascii_uppercase) for _ in range(length))


# Copyright Okera Inc.

from __future__ import absolute_import

import csv
import datetime
import io

from okera._thrift_api import (TTypeId)

def __gen_random_data(t, idx):
    if t.type.type_id == TTypeId.STRING or t.type.type_id == TTypeId.VARCHAR:
        return "string_%d" % idx
    elif t.type.type_id == TTypeId.CHAR:
        return "char_%d" % idx
    elif t.type.type_id == TTypeId.BOOLEAN:
        if idx % 2 == 0:
            return True
        return False
    elif t.type.type_id == TTypeId.TINYINT:
        return idx
    elif t.type.type_id == TTypeId.SMALLINT:
        return idx
    elif t.type.type_id == TTypeId.INT:
        return idx
    elif t.type.type_id == TTypeId.BIGINT:
        return idx
    elif t.type.type_id == TTypeId.FLOAT:
        return idx
    elif t.type.type_id == TTypeId.DOUBLE:
        return idx
    elif t.type.type_id == TTypeId.DATE:
        epoch = datetime.datetime(1970, 1, 1, 0, 0)
        return (epoch + datetime.timedelta(idx)).date()
    elif t.type.type_id == TTypeId.TIMESTAMP_NANOS:
        return datetime.datetime.fromtimestamp(idx)
    elif t.type.type_id == TTypeId.DECIMAL:
      return idx
    else:
        raise RuntimeError(
            "Unsupported type: " + TTypeId._VALUES_TO_NAMES[t.type.type_id])

def gen_random_data(conn, db, name, num_records=10):
    """ Generates random data data that matches the schema for this dataset.

        Parameters
        ----------
        conn : PlannerConnection
            Connection to a planner.

        db : str
            Database containing this dataset.

        name : str
            Name of this dataset.

        num_records : int, optional
            Number of records to generate. Default is 10.

        Returns
        -------
        list(list(object))
            Returns the result as a table.
    """
    datasets = conn.list_datasets(db, name=name)
    if not datasets:
        raise RuntimeError("Invalid dataset: %s.%s" % (db, name))
    schema = datasets[0].schema
    data = []
    for i in range(0, num_records):
        row = []
        for col in schema.cols:
            row.append(__gen_random_data(col, i))
        data.append(row)
    return data

def gen_random_csv(conn, db, name, num_records=10):
    """ Generates random CSV formatted data that matches the schema for this dataset.

        Parameters
        ----------
        conn : PlannerConnection
            Connection to a planner.

        db : str
            Database containing this dataset.

        name : str
            Name of this dataset.

        num_records : int, optional
            Number of records to generate. Default is 10.

        Returns
        -------
        str
            Returns generated CSV.
    """
    data = gen_random_data(conn, db, name, num_records)
    output = io.StringIO()
    writer = csv.writer(output)
    writer.writerows(data)
    return output.getvalue().strip()




# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Tests that should run on any configuration. The server auth can be specified
# as an environment variables before running this test.
# pylint: disable=unused-variable, unused-import, bad-indentation, no-self-use
import time
import unittest
import requests
from okera import (HAS_PANDAS, HAS_NUMPY, NO_PANDAS_RUNTIME_ERROR, NO_NUMPY_RUNTIME_ERROR)
from okera.tests import pycerebro_test_common as common

class DependencyTests(unittest.TestCase):

  def test_no_pandas_assert(self):
    ctx = common.get_test_context()
    with common.get_planner(ctx) as planner:
      HAS_NUMPY = True
      HAS_PANDAS = False
      try:
        planner.scan_as_pandas("okera_sample.sample")
      except (RuntimeError) as e:
        self.assertEqual(NO_PANDAS_RUNTIME_ERROR,
                         str(e),
                         msg='Expected NO_PANDAS_RUNTIME_ERROR but received other')
      planner.scan_as_json("okera_sample.sample")

  def test_no_numpy_assert(self):
    ctx = common.get_test_context()
    with common.get_planner(ctx) as planner:
      HAS_NUMPY = False
      HAS_PANDAS = False # Pandas requires numpy
      try:
        planner.scan_as_json("okera_sample.sample")
      except (RuntimeError) as e:
        self.assertEqual(NO_NUMPY_RUNTIME_ERROR,
                         str(e),
                         msg='Expected NO_NUMPY_RUNTIME_ERROR but received other')

      try:
        planner.scan_as_pandas("okera_sample.sample")
      except (RuntimeError) as e:
        self.assertEqual(NO_PANDAS_RUNTIME_ERROR,
                         str(e),
                         msg='Expected NO_PANDAS_RUNTIME_ERROR but received other')

  def test_no_numpy_presto(self):
    ctx = common.get_test_context()
    with common.get_planner(ctx, dialect='presto') as planner:
      HAS_NUMPY = False
      HAS_PANDAS = False
      planner.scan_as_json("select * from okera_sample.sample")

if __name__ == "__main__":
  unittest.main()



# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for Presto in PyOkera
#
# pylint: disable=broad-except
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-locals
# pylint: disable=bad-continuation
# pylint: disable=broad-except

import unittest
import os
import pytest

import urllib3
import prestodb

from okera._thrift_api import TRecordServiceException
from okera.tests import pycerebro_test_common as common

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

DB = 'test_views_db'
VIEW = 'testview'

DEFAULT_PRESTO_PORT = int(os.environ['ODAS_TEST_PORT_PRESTO_COORD_HTTPS'])
BAD_PRESTO_PORT = 1234

def create_view(conn, db, name, cols, external, skip_analysis, stmt):
    view_stmt = ''
    if external:
        view_stmt = 'CREATE EXTERNAL VIEW %s.%s ' % (db, name)
    else:
        view_stmt = 'CREATE VIEW %s.%s ' % (db, name)

    if cols:
        view_stmt += '('
        view_stmt += ', '.join(['%s %s' % (c[0], c[1]) for c in cols])
        view_stmt += ') '

    if skip_analysis:
        view_stmt += 'SKIP_ANALYSIS '

    view_stmt += 'AS %s' % stmt

    conn.execute_ddl(view_stmt)

def create_view_with_data(conn, db, name, cols, stmt):
    view_stmt = 'CREATE EXTERNAL VIEW %s.%s ' % (db, name)

    if cols:
        view_stmt += '('
        view_stmt += ', '.join(['%s %s' % (c[0], c[1]) for c in cols])
        view_stmt += ') '

    view_stmt += 'SKIP_ANALYSIS '

    view_stmt += 'USING VIEW DATA AS "%s"' % stmt

    conn.execute_ddl(view_stmt)

def cleanup(conn, db):
    conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db)
    conn.execute_ddl('CREATE DATABASE %s' % db)

@pytest.mark.filterwarnings('ignore::urllib3.exceptions.InsecureRequestWarning')
@pytest.mark.filterwarnings('ignore:numpy.dtype size changed')
class ViewsTest(unittest.TestCase):
    def test_create_with_skip_no_external(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            cleanup(conn, DB)
            with pytest.raises(TRecordServiceException):
                create_view(
                    conn, DB, VIEW,
                    [('x', 'INT'), ('a', 'INT')],
                    external=False, skip_analysis=True,
                    stmt='select * from okera_sample.whoami')

    def test_create_with_skip_no_columns(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            cleanup(conn, DB)
            with pytest.raises(TRecordServiceException):
                create_view(
                    conn, DB, VIEW,
                    [],
                    external=True, skip_analysis=True,
                    stmt='select * from okera_sample.whoami')

    def test_create_with_skip_no_types(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            cleanup(conn, DB)
            with pytest.raises(TRecordServiceException):
                create_view(
                    conn, DB, VIEW,
                    [('x', ''), ('a', '')],
                    external=True, skip_analysis=True,
                    stmt='select * from okera_sample.whoami')

    def test_create_with_skip_table_no_exist(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            cleanup(conn, DB)
            create_view(
                conn, DB, VIEW,
                [('c1', 'int'), ('c2', 'int')],
                external=True, skip_analysis=True,
                stmt='select * from doesnot.exist')
            with pytest.raises(prestodb.exceptions.PrestoQueryError):
                conn.scan_as_json(
                    'select * from %s.%s' % (DB, VIEW), dialect='presto')

    def test_create_with_skip_stale(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            cleanup(conn, DB)
            create_view(
                conn, DB, VIEW,
                [('record', 'string'), ('c2', 'int')],
                external=True, skip_analysis=True,
                stmt='select record, 1 from okera_sample.sample')
            with pytest.raises(prestodb.exceptions.PrestoQueryError):
                conn.scan_as_json(
                    'select * from %s.%s' % (DB, VIEW), dialect='presto')

    def test_create_with_skip_working(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            cleanup(conn, DB)
            create_view(
                conn, DB, VIEW,
                [('record', 'string'), ('c2', 'int')],
                external=True, skip_analysis=True,
                stmt='select record, 1 as c2 from okera_sample.sample')
            conn.scan_as_json(
                'select * from %s.%s' % (DB, VIEW), dialect='presto')

    def test_create_with_skip_presto_functions(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            cleanup(conn, DB)
            create_view(
                conn, DB, VIEW,
                [('dob_parts', 'array<string>'), ('dob', 'string')],
                external=True, skip_analysis=True,
                stmt='''select
                            split(dob, "-") as dob_parts, dob
                        from okera_sample.users
                        limit 10''')
            rows = conn.scan_as_json(
                'select * from %s.%s' % (DB, VIEW), dialect='presto')
            for row in rows:
                assert row['dob'] == '-'.join(row['dob_parts'])

    def test_create_view_data(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            cleanup(conn, DB)
            stmt = 'not a query'
            create_view_with_data(
                conn, DB, VIEW,
                [('dob_parts', 'array<string>'), ('dob', 'string')],
                stmt=stmt)
            views = conn.list_datasets(db=DB, name=VIEW)
            assert len(views) == 1
            assert views[0].view_expanded_text == stmt



import logging

from okera.integration import boto as okera_boto

LOG = logging.getLogger(__name__)

_VALID_COMMANDS = ['s3', 's3api']
_CONFIG_KEY_OKERA = 'okera'
_CONFIG_KEY_REST = 'rest'
_CONFIG_KEY_PROXY = 'proxy'
_CONFIG_KEY_TOKEN = 'token'

def _get_info_from_profile(profile):
    proxy_endpoint = None
    rest_endpoint = None
    token = None

    if _CONFIG_KEY_OKERA not in profile:
        return None

    info = profile[_CONFIG_KEY_OKERA]
    if _CONFIG_KEY_PROXY not in info \
        or _CONFIG_KEY_REST not in info \
        or _CONFIG_KEY_TOKEN not in info:
        return None

    proxy_endpoint = profile[_CONFIG_KEY_OKERA][_CONFIG_KEY_PROXY]
    rest_endpoint = profile[_CONFIG_KEY_OKERA][_CONFIG_KEY_REST]
    token = profile[_CONFIG_KEY_OKERA][_CONFIG_KEY_TOKEN]

    return proxy_endpoint, rest_endpoint, token

def _register_okera_proxy(parsed_args, **kwargs):
    if parsed_args.debug:
        LOG.setLevel(logging.DEBUG)

    if not parsed_args.command:
        LOG.debug("No command")
        return

    command = parsed_args.command.lower()
    if command not in _VALID_COMMANDS:
        LOG.debug("Not valid command: %s" % command)
        return

    if 'session' not in kwargs or not kwargs['session']:
        LOG.debug("No session")
        return

    session = kwargs['session']
    if parsed_args.profile:
        session.set_config_variable('profile', parsed_args.profile)

    okera_info = _get_info_from_profile(session.get_scoped_config())
    if not okera_info:
        LOG.debug("No Okera configuration found")
        return

    LOG.debug("Injecting Okera proxy")
    proxy, rest, token = okera_info

    okera_session = okera_boto.okera_session(session, token, rest, proxy)
    kwargs['session'] = okera_session

    # TODO: rather than setting this, we should automatically
    # retrieve the CA bundle from the server and set it so we
    # still validate it.
    parsed_args.verify_ssl = False

def awscli_initialize(cli):
    cli.register('top-level-args-parsed', _register_okera_proxy)


# Copyright 2020 Okera Inc. All Rights Reserved.
#
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=no-else-return

import unittest
import io
import json
import os
import warnings
import PIL.Image as Image
import PyPDF2

from okera.tests import pycerebro_test_common as common

DEFAULT_PRESTO_PORT = int(os.environ['ODAS_TEST_PORT_PRESTO_COORD_HTTPS'])

## The all types check here will need to be verified against the JDBC data source,
## maybe by running query in DBeaver.
class JdbcScanTest(unittest.TestCase):

    def format(self, data):
        return json.dumps(data, sort_keys=True, indent=1)

    def assert_output(self, qry, expected, is_scan=True):
        warnings.filterwarnings("ignore", message="numpy.dtype size changed")
        ctx = common.get_test_context(dialect='presto')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            if is_scan:
                result = conn.scan_as_json(qry, strings_as_utf8=False)
            else:
                result = conn.execute_ddl(qry)
            print("Actual and expected: ")
            print(format(result))
            print(expected)
            print("Done printing.....")
            assert result == expected

    def test_dremio_basic(self):
        CXN = 'test_dremio_connection'
        DB = 'test_dremio'
        CONFIG = 's3://cerebrodata-dev/jdbc/dremio.properties'
        ctx = common.get_test_context()

        create_connection = [
            """
                CREATE DATACONNECTION %s CXNPROPERTIES
                (
                'connection_type'='JDBC',
                'jdbc_driver'='dremio',
                'user_key'='awssm://dremio_internal_akshay',
                'password_key'='awssm://dremio_internal_akshay_password',
                'host'='dremio.internal.okera.rocks',
                'port'='31010',
                'connection_properties'='{"driver.jar.path":"s3://okera-jdbc-test/drivers/dremio-jdbc-driver-12.0.0.jar", "fetchSize": "64"}'
                )
            """ % CXN,

            # Using connection pool
            """
                CREATE DATACONNECTION %s CXNPROPERTIES
                (
                'connection_type'='JDBC',
                'jdbc_driver'='dremio',
                'user_key'='awssm://dremio_internal_akshay',
                'password_key'='awssm://dremio_internal_akshay_password',
                'host'='dremio.internal.okera.rocks',
                'port'='31010',
                'connection_properties'='{"driver.jar.path":"s3://okera-jdbc-test/drivers/dremio-jdbc-driver-12.0.0.jar", "fetchSize": "64", "supports.connection-pool"="true"}'
                )
            """ % CXN,

            # TODO: add following variants:
            # 1. Using newer driver
            # 2. Using built in driver (with and without pooling)
        ]

        with common.get_planner(ctx) as conn:
            # The 'test' db exists in this dremio cluster but is empty
            for create in create_connection:
                conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % DB)
                conn.execute_ddl('DROP DATACONNECTION IF EXISTS %s' % CXN)
                conn.execute_ddl(create)
                conn.execute_ddl("""CREATE DATABASE %s DBPROPERTIES(
                    'okera.connection.name' = '%s',
                    'jdbc.schema.name'='okera.okera-demo',
                    'okera.autotagger.skip'='true')""" % (DB, CXN))
                conn.execute_ddl('ALTER DATABASE %s LOAD DEFINITIONS()' % DB)
                datasets = conn.list_dataset_names(DB)
                self.assertTrue(('%s.hospital_discharge_sampled' % DB) in datasets,
                                msg=str(datasets))
                result = conn.scan_as_json(
                    'select facility_name from %s.hospital_discharge_sampled limit 1' %\
                    DB)
                print(result)
                self.assertEqual(len(result), 1)
                self.assertEqual(result[0]['facility_name'], 'Oneida Healthcare Center')

                conn.scan_as_json('%s.hospital_discharge_sampled limit 250' % DB)

    def test_jdbc_blob_img_file_download(self):
        qry = "select * from jdbc_test_oracle.blob_test where id = 1"
        warnings.filterwarnings("ignore", message="numpy.dtype size changed")
        ctx = common.get_test_context(dialect='okera')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            result = conn.scan_as_pandas(qry)
        data = result.to_dict('records')

        # if underlying "blob_data" column is null, it is returned as float.
        # For non null values, data should be returned as bytes. That's why below assert
        self.assertTrue(isinstance(data[0]['blob_data'], bytes), \
            "The value of 'blob_data' column is not returned as bytes!! ")

        # Cleanup if file exists.
        tmp_image_file = '/tmp/test_jdbc_blob_img_file_download.png'
        if os.path.exists(tmp_image_file):
            os.remove(tmp_image_file)

        # Download image and save as png. We know the original one was png format.
        image = Image.open(io.BytesIO(data[0]['blob_data']))
        image.save(tmp_image_file)

        # Open and verify the image.
        if os.path.exists(tmp_image_file):
            image = Image.open(tmp_image_file)
            # Assert the image size.
            assert image.size[0] == 1000
            assert image.size[1] == 562
            image.verify()
        else:
            # no file is downloaded. underlying jdbc source might have 'blob_data' as NULL
            self.assertTrue(1 == 2, "No image file availalbe!! Please investigate.")

    def test_jdbc_blob_pdf_file_download(self):
        qry = "select * from jdbc_test_oracle.blob_test where id = 2"
        warnings.filterwarnings("ignore", message="numpy.dtype size changed")
        ctx = common.get_test_context(dialect='okera')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            result = conn.scan_as_pandas(qry)
        data = result.to_dict('records')

        # if underlying "blob_data" column is null, it is returned as float.
        # For non null values, data should be returned as bytes. That's why below assert
        self.assertTrue(isinstance(data[0]['blob_data'], bytes), \
            "The value of 'blob_data' column is not returned as bytes!! ")

        # Cleanup if file exists.
        tmp_pdf_file = '/tmp/test_jdbc_blob_pdf_file_download.pdf'
        if os.path.exists(tmp_pdf_file):
            os.remove(tmp_pdf_file)

        # Read the pdf contents which is stored as blob
        pdf_reader = PyPDF2.PdfFileReader(io.BytesIO(data[0]['blob_data']), strict=False)
        assert pdf_reader.numPages == 2

        # Write pdf contents into the blank file "test_jdbc_blob_pdf_file_download.pdf"
        pdf_writer = PyPDF2.PdfFileWriter()
        for page_num in range(pdf_reader.numPages):
            page_obj = pdf_reader.getPage(page_num)
            pdf_writer.addPage(page_obj)
        pdf_output_file = open(tmp_pdf_file, 'wb')
        pdf_writer.write(pdf_output_file)
        pdf_output_file.close()

        # Assert the no. of pages in the newly written file
        pdf_reader = PyPDF2.PdfFileReader(tmp_pdf_file)
        assert pdf_reader.numPages == 2

    def test_jdbc_clob_txt_file_download(self):
        qry = "select * from jdbc_test_oracle.clob_test where id = 1"
        warnings.filterwarnings("ignore", message="numpy.dtype size changed")
        ctx = common.get_test_context(dialect='okera')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            result = conn.scan_as_pandas(qry)
        data = result.to_dict('records')

        # if underlying "clob_data" column is null, it is returned as float.
        # For non null values, data should be returned as bytes. That's why below assert
        self.assertTrue(isinstance(data[0]['clob_data'], bytes), \
            "The value of 'clob_data' column is not returned as bytes!! ")

        # Cleanup if file exists.
        tmp_txt_file = '/tmp/test_jdbc_clob_txt_file_download.txt'
        file_size = 0
        if os.path.exists(tmp_txt_file):
            os.remove(tmp_txt_file)

        # Open the blank file, write the data and verify the file size
        with open(tmp_txt_file, 'wb') as file_handle:
            file_handle.write(bytes(data[0]['clob_data']))
            file_handle.close()
            file_size = os.path.getsize(tmp_txt_file)

        # Assert the file size (in bytes)
        assert file_size == 6121

    def test_jdbc_all_types(self):
        self.assert_output(
            'select * from jdbc_test_mysql.all_types_v2',
            [{'date': None, 'decimal': None, 'float': None, 'mediumint': None,
              'blob': None, 'tinyblob': None, 'bigint': None,
              'mediumtext': None, 'datetime': None, 'time': None,
              'longtext': None, 'mediumblob': None, 'smallint': None, 'enum': None,
              'varchar': None, 'tinytext': None, 'timestamp': None, 'binary': None,
              'longblob': None, 'int': None, 'set': None, 'tinyint': None, 'bool': None,
              'varbinary': None, 'text': None, 'char': None, 'double': None},
             {'date': '1990-01-01', 'decimal': '8.23', 'float': 7.099999904632568,
              'mediumint': 3, 'blob': 'blob', 'tinyblob': 'tinyblob', 'bigint': 5,
              'mediumtext': 'mediumtext', 'datetime': '2020-10-01 18:19:03.000',
              'time': '1970-01-01 00:20:20.000', 'longtext': 'longtext',
              'mediumblob': 'mediumblob', 'smallint': 2, 'enum': '1', 'varchar':
              'test', 'tinytext': 'tinytext', 'timestamp': '2017-06-01 18:19:03.000',
              'binary': 'binary\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',
              'longblob': 'longblob', 'int': 4, 'set': '1,2  ', 'tinyint': 1,
              'bool': True, 'varbinary': 'varbinary', 'text': 'hello',
              'char': 'char      ', 'double': 6.0}])

        self.assert_output(
            'select * from jdbc_test_psql.all_types',
            [{'bigint': None, 'bigserial': 1, 'money': None, 'numeric': None,
              'text': None, 'timestamp': None, 'real': None, 'serial': 1, 'bool': None,
              'int': None, 'double': None, 'time': None, 'smallint': None, 'bit': None,
              'varchar': None, 'decimal': None, 'char': None},
             {'bigint': 3, 'bigserial': 2, 'money': 10.0, 'numeric': '7.10',
              'text': 'hello', 'timestamp': '2017-06-01 18:19:03.000', 'real': 2.13000011,
              'serial': 2, 'bool': True, 'int': 2, 'double': 6.0,
              'time': '1970-01-01 18:19:03.000', 'smallint': 1,
              'bit': 1010101010, 'varchar': 'test', 'decimal': '8.23',
              'char': 'char      '}])

        self.assert_output(
            'select * from jdbc_test_redshift.all_types',
            [{'int': None, 'decimal': None, 'text': None, 'smallint': None,
              'numeric': None, 'real': None, 'char': None, 'double': None, 'bool': None,
              'varchar': None, 'timestamp': None, 'bigint': None},
             {'int': 2, 'decimal': '8.23', 'text': 'hello', 'smallint': 1,
              'numeric': '7.10', 'real': 10.0, 'char': 'char      ',
              'double': 6.0, 'bool': True, 'varchar': 'test',
              'timestamp': '2017-06-01 18:19:03.000', 'bigint': 3}])

        self.assert_output(
            'select * from jdbc_test_oracle.all_types',
            [{'int_col': 4, 'date_col': '2017-06-01 00:00:00.000',
              'timestamp_col': '2017-06-01 18:19:03.000', 'smallint_col': 3,
              'dec_col': '4000000.45', 'number_col': '1.000000', 'numeric_col': 2,
              'bigint_col': '100000000000', 'decimal_col': '3000.45',
              'char_col': 'char      ', 'float_col': 3.4000000953674316,
              'varchar_col': 'varchar2', 'nvarchar_col': 'nvarchar2'},
             {'int_col': None, 'date_col': None, 'timestamp_col': None,
              'smallint_col': None, 'dec_col': None, 'number_col': None,
              'numeric_col': None, 'bigint_col': None, 'decimal_col': None,
              'char_col': None, 'float_col': None, 'varchar_col': None,
              'nvarchar_col': None}])

        ## TODO for snowflake, athena and sqlserver

    def check_jdbc_source_by_schema(self, **kwargs):
        '''
            A common test method to create a jdbc_data_source by schema and iterate
            through the tables and verify the query is successful.
        '''
        # Eventually, we can just call test_authorization on this.
        select_queries = [
            'select * from {}',
            'select count(*) from {}'
            ]
        ctx = common.get_test_context(dialect='okera')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx) as conn:
            db_name = kwargs['db_name']
            schema_name = kwargs['schema_name']
            creds_file = kwargs['credentials_file']
            # Cleanup existing DB
            conn.execute_ddl("DROP DATABASE IF EXISTS {} CASCADE".format(db_name))
            # Create the test database with credentials file.
            conn.execute_ddl('''
                CREATE DATABASE IF NOT EXISTS {} DBPROPERTIES
                ('credentials.file' = '{}',
                 'jdbc.schema.name'='{}')'''.format(db_name, creds_file, schema_name))
            # Prepare datasets
            conn.execute_ddl("ALTER DATABASE {} LOAD DEFINITIONS()".format(db_name))

            res = conn.execute_ddl("SHOW TABLES IN {} ".format(db_name))
            print(res)
            print(len(res))
            self.assertEqual(len(res), kwargs['expected_table_count'])
            if kwargs['skip_scan']:
                return
            for table in res:
                if table[0] in kwargs['ignore_tables_list']:
                    continue
                for query in select_queries:
                    result = conn.scan_as_pandas(query.format(db_name + "." + table[0]))
                    # dummy statement
                    print(len(result))

    @unittest.skip("SAP HANA instance can be costly to maintain, skip tests for normal.")
    def test_saphana_jdbc_source(self):
        '''
            This test is run on-demand due to maintenance cost of SAP HANA instances.
            If needed to test, a standard SAP HANA Express Edition 2.0 can be spun up
            and we check for some schemas in the pre-loaded SAP HANA instances.
        '''
        # These schemas have some interesting tables/views.
        # SAP HANA has some special views like CALC VIEW, HIERARCHY VIEW etc which are
        # based on stored procedures. No JDBC tools support it (like DBeaver etc), we
        # skip those views for scanning. The customer has a choice to include them
        # by setting `table.types` property in the credentials file.
        # See 's3://cerebro-datasets/jdbc_demo/jdbc_saphana_all_table_types.conf'.
        # The HANA_XS_BASE and _SI_BIC schema is full of such special views.
        # Those two schemas also have very special characters which are good test cases.
        # The _SYS_STATISTICS is a big schema and good stress test for metadata operation.
        saphana_schemas_to_verify = [
            ('_SYS_BI', 57, False, ['get_bwauths4hana_tabletype_valtab']),
            ('HANA_XS_BASE', 23, True, []),
            ('_SYS_BIC', 10, True, []),
            ('_SYS_STATISTICS', 388, False,
             ['tt_mail_collector', 'tt_statistics_alert_thresholds',
              'tt_statistics_used_values'])]
        cred_file = 's3://cerebro-datasets/jdbc_demo/jdbc_saphana_all_table_types.conf'
        for schema in saphana_schemas_to_verify:
            (s_name, count, skip, ignore_tables) = schema
            self.check_jdbc_source_by_schema(
                db_name='jdbc_saphana_test', credentials_file=cred_file,
                schema_name=s_name, expected_table_count=count, skip_scan=skip,
                ignore_tables_list=ignore_tables)



# Copyright Okera Inc.
#
#
#
# pylint: disable=wrong-import-order
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

try:
  import numpy
  HAS_NUMPY = True
except ImportError:
  HAS_NUMPY = False

try:
  import pandas
  HAS_PANDAS = True
except ImportError:
  HAS_PANDAS = False

NO_PANDAS_RUNTIME_ERROR = "The pandas library is required to perform scan operations."
NO_NUMPY_RUNTIME_ERROR = "The numpy library is required to perform scan operations."

def assert_numpy_installed():
  if not HAS_NUMPY:
    raise RuntimeError(NO_NUMPY_RUNTIME_ERROR)

def assert_pandas_installed():
  if not HAS_PANDAS:
    raise RuntimeError(NO_PANDAS_RUNTIME_ERROR)

def should_patch_botocore():
  import os
  if os.environ.get('OKERA_PATCH_BOTO', None):
    return True
  return False

def check_and_patch_botocore():
  if should_patch_botocore():
    from okera.botocore_patch import patch_botocore
    patch_botocore()

# Import the public API
from okera.odas import context, version
_default_app_context = context()

def initialize_default_context():
  import os
  host = os.environ.get('OKERA_PLANNER_HOST', None)
  if host:
    port = int(os.environ.get('OKERA_PLANNER_PORT', 12050))
    token = os.environ.get('OKERA_USER_TOKEN', None)
    if token:
      _default_app_context.enable_token_auth(token_str=token)
    return _default_app_context.connect(host=host, port=port)

_default_context = None
def get_default_context():
  global _default_context
  if not _default_context:
    _default_context = initialize_default_context()
  return _default_context

# Initialize boto3 patches if necessary
check_and_patch_botocore()

from okera.integration.aws import awscli_initialize